from dpks.quant_matrix import QuantMatrix
import pytest
import xgboost
from sklearn.svm import SVC
from sklearn import tree
from sklearn.neighbors import KNeighborsClassifier


@pytest.fixture
def quantified_data(paths):
    qm = QuantMatrix(
        quantification_file=str(paths["sepsis_matrix_path"]),
        design_matrix_file=str(paths["sepsis_design_path"]),
    )
    quantified_data = qm.normalize(
        method="mean",
    ).quantify(method="top_n", summarization_method="mean")
    return quantified_data


def test_xgb(quantified_data):
    clf = xgboost.XGBClassifier()
    trained_classifier = quantified_data.train(clf)
    quantified_data.interpret(trained_classifier.esimator_, trained_classifier.scaler)


def test_decision_tree(quantified_data):
    clf = tree.DecisionTreeClassifier()
    trained_classifier = quantified_data.train(clf)
    quantified_data.interpret(
        trained_classifier.esimator_,
        trained_classifier.scaler,
        shap_algorithm="tree",
    )


def test_knn(quantified_data):
    clf = KNeighborsClassifier()
    trained_classifier = quantified_data.train(clf)
    quantified_data.interpret(
        trained_classifier.esimator_,
        trained_classifier.scaler,
        shap_algorithm="permutation",
    )


def test_svm(quantified_data):
    clf = SVC()
    trained_classifier = quantified_data.train(clf)
    quantified_data.interpret(
        trained_classifier.esimator_,
        trained_classifier.scaler,
        shap_algorithm="linear",
    )



from typing import Union

import numpy as np
import pandas as pd


def parse_diann(
    diann_file: Union[str, pd.DataFrame], diann_qvalue: float = 0.01
) -> pd.DataFrame:
    if isinstance(diann_file, str):
        quantification_file = pd.read_csv(diann_file, sep="\t")

    else:
        quantification_file = diann_file

    sample_q_value_column = "Q.Value"

    if "Lib.Q.Value" in quantification_file:
        sample_q_value_column = "Lib.Q.Value"

    quantification_file = quantification_file[
        quantification_file[sample_q_value_column] <= diann_qvalue
    ]

    used_columns = []

    scoring_columns = []

    if ("Lib.PG.Q.Value" in quantification_file) or (
        "Lib.Q.Value" in quantification_file
    ):
        if "Lib.PG.Q.Value" not in quantification_file:
            quantification_file["Lib.PG.Q.Value"] = 0.0

        columns = [
            "Protein.Ids",
            # "Genes",
            "Precursor.Id",
            "Modified.Sequence",
            "Precursor.Charge",
            "RT",
            "Precursor.Quantity",
            "Lib.PG.Q.Value",
            "Lib.Q.Value",
        ]

        scoring_columns = ["Lib.PG.Q.Value", "Lib.Q.Value"]

        for col in columns:
            if col in quantification_file:
                used_columns.append(col)

    else:
        columns = [
            "Protein.Ids",
            # "Genes",
            "Precursor.Id",
            "Modified.Sequence",
            "Precursor.Charge",
            "RT",
            "Precursor.Quantity",
            "Global.PG.Q.Value",
            "Global.Q.Value",
        ]

        scoring_columns = ["Global.PG.Q.Value", "Global.Q.Value"]

        for col in columns:
            if col in quantification_file:
                used_columns.append(col)

    file_column = "Run"

    if "Run" in quantification_file:
        long_results = quantification_file[["Run"] + used_columns].copy()

    elif "File.Name" in quantification_file:
        file_column = "File.Name"

        long_results = quantification_file[["File.Name"] + used_columns].copy()

    # Removes file extension (mzml, d, etc) and then rebuilds sample name
    long_results[file_column] = long_results[file_column].str.rsplit(".").str[0]

    retention_times = (
        long_results[
            [
                "Protein.Ids",
                # "Genes",
                "Precursor.Id",
                "Modified.Sequence",
                "Precursor.Charge",
                "RT",
            ]
        ]
        .groupby(
            [
                "Protein.Ids",
                # "Genes",
                "Precursor.Id",
                "Modified.Sequence",
                "Precursor.Charge",
            ]
        )
        .agg("median")
    )

    wide_results = (
        long_results.pivot(
            columns=file_column,
            index=[
                "Precursor.Id",
                "Protein.Ids",
                # "Genes",
                "Modified.Sequence",
                "Precursor.Charge",
            ]
            + scoring_columns,
            values="Precursor.Quantity",
        )
        .reset_index()
        .copy()
    )

    wide_results.columns.name = None

    wide_results = (
        wide_results.set_index(
            [
                "Protein.Ids",
                # "Genes",
                "Precursor.Id",
                "Modified.Sequence",
                "Precursor.Charge",
            ]
        )
        .join(retention_times)
        .reset_index()
    )

    if ("Lib.PG.Q.Value" in quantification_file) or (
        "Lib.Q.Value" in quantification_file
    ):
        wide_results.rename(
            columns={
                "Precursor.Id": "PrecursorId",
                "Protein.Ids": "Protein",
                # "Genes": "ProteinLabel",
                "Modified.Sequence": "PeptideSequence",
                "Precursor.Charge": "Charge",
                "Lib.PG.Q.Value": "ProteinQValue",
                "Lib.Q.Value": "PeptideQValue",
                "RT": "RetentionTime",
            },
            inplace=True,
        )

    else:
        wide_results.rename(
            columns={
                "Precursor.Id": "PrecursorId",
                "Protein.Ids": "Protein",
                # "Genes": "ProteinLabel",
                "Modified.Sequence": "PeptideSequence",
                "Precursor.Charge": "Charge",
                "Global.PG.Q.Value": "ProteinQValue",
                "Global.Q.Value": "PeptideQValue",
                "RT": "RetentionTime",
            },
            inplace=True,
        )

    wide_results.replace(0, np.nan, inplace=True)

    return wide_results



from typing import Optional, List

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.feature_selection import RFE
from sklearn.model_selection import (
    cross_val_score,
)
from sklearn.utils import resample


class BoostrapRFE:
    """
    Bootstrap Recursive Feature Elimination (RFE) for feature selection.

    Parameters:
    - min_features_to_select (int): Minimum number of features to select.
    - step (int): Step size for RFE.
    - importance_getter (str): Method to get feature importances.
    - scoring (str): Scoring metric for feature selection.
    - k_folds (int): Number of folds for cross-validation.
    - threads (int): Number of threads for parallel processing.
    - verbose (bool): Verbosity of the process.
    - shap_algorithm (str): SHAP algorithm to use.
    - random_state (int): Random seed for reproducibility.
    - shuffle (bool): Whether to shuffle data during resampling.
    - replace (bool): Whether to use replacement during resampling.
    - downsample_rate (float): Rate of downsampling during resampling.
    - feature_names (Optional[List[str]]): List of feature names.

    Methods:
    - fit(X, y, classifier): Fit the BootstrapRFE instance to the data.
    - get_scores(): Get the scores of feature selectors.
    - get_ranks(): Get the rankings of selected features.
     - evaluate(X, y, classifier): Evaluate the performance with feature subsets.

    """

    def __init__(
        self,
        min_features_to_select: int = 10,
        step: int = 3,
        importance_getter: str = "auto",
        scoring: str = "accuracy",
        k_folds: int = 3,
        threads: int = 1,
        verbose: bool = False,
        shap_algorithm: str = "auto",
        random_state: int = 42,
        shuffle: bool = True,
        replace: bool = True,
        downsample_rate: float = 0.8,
        feature_names: Optional[List[str]] = None,
    ) -> None:
        """
        Initialize the BoostrapRFE instance with specified parameters.

        Parameters:
        ...

        Returns:
        None
        """
        self.selectors = dict()
        self.results = dict()
        self.verbose = verbose
        self.threads = threads
        self.scoring = scoring
        self.k_folds = k_folds
        self.models = dict()
        self.min_features_to_select = min_features_to_select
        self.step = step
        self.importance_getter = importance_getter
        self.shap_algorithm = shap_algorithm
        self.random_state = random_state
        self.shuffle = shuffle
        self.replace = replace
        self.downsample_rate = downsample_rate
        self.feature_names = feature_names

    def fit(self, X, y, classifier) -> None:
        """
        Fit the BootstrapRFE instance to the data.

        Parameters:
        - X: Input data.
        - y: Target values.
        - classifier: Classifier for feature selection.

        Returns:
        None
        """

        self.results = Parallel(n_jobs=self.threads)(
            delayed(_rank_features)(
                classifier,
                X,
                y,
                self.scoring,
                self.replace,
                self.downsample_rate,
                self.step,
                self.min_features_to_select,
                self.importance_getter,
            )
            for i in range(self.k_folds)
        )

    def get_scores(self):
        """
        Get the scores of feature selectors.

        Returns:
        pd.DataFrame: DataFrame containing feature scores.
        """

        selector_scores = []
        feature_number = []
        indices = []

        for idx, (_, scores) in enumerate(self.results):
            for score in scores:
                selector_scores.append(scores[score])
                feature_number.append(score)
                indices.append(idx)

        scores = pd.DataFrame(
            {"idx": indices, "feature_number": feature_number, "score": selector_scores}
        )

        return scores

    def get_ranks(self):
        """
        Get the rankings of selected features.

        Returns:
        pd.DataFrame: DataFrame containing feature rankings.
        """

        selector_rankings = dict()

        scores = self.get_scores()

        for idx, (selector, _) in enumerate(self.results):
            selector_rankings[idx] = selector.ranking_

        feature_ranks = pd.DataFrame(
            {
                "feature_name": self.feature_names,
            }
        )

        for idx, ranking in selector_rankings.items():
            idx_scores = scores[scores["idx"] == idx][["feature_number", "score"]]

            feature_ranks[f"ranking_{idx}"] = ranking

            idx_scores = idx_scores.rename(columns={"score": f"score_{idx}"})

            feature_ranks = feature_ranks.join(
                idx_scores.set_index("feature_number"), on=f"ranking_{idx}"
            )

        feature_ranks = feature_ranks.copy()

        feature_ranks["feature_rank_mean"] = feature_ranks[
            [col for col in feature_ranks.columns if "ranking_" in col]
        ].mean(axis=1)
        feature_ranks["feature_rank_std"] = feature_ranks[
            [col for col in feature_ranks.columns if "ranking_" in col]
        ].std(axis=1)
        feature_ranks["feature_rank_median"] = feature_ranks[
            [col for col in feature_ranks.columns if "ranking_" in col]
        ].median(axis=1)

        feature_ranks["adjusted_rank"] = (
            feature_ranks["feature_rank_median"] + feature_ranks["feature_rank_std"]
        )

        return feature_ranks

    def evaluate(self, X, y, classifier):
        """
        Evaluate the performance with feature subsets.

        Parameters:
        - X: Input data.
        - y: Target values.
        - classifier: Classifier for feature selection.

        Returns:
        pd.DataFrame: DataFrame containing evaluation results.
        """

        rank_values = []
        rank_scores = []

        feature_ranks = self.get_ranks()

        for rank_value in feature_ranks.sort_values("adjusted_rank", ascending=False)[
            "adjusted_rank"
        ].values:
            if isinstance(X, pd.DataFrame):
                selected_proteins = feature_ranks[
                    feature_ranks["adjusted_rank"] <= rank_value
                ]["feature_name"].values

                X_subset = X[selected_proteins]

            ## Need numpy array subsetting here also

            scores = cross_val_score(classifier, X_subset, y, scoring="accuracy", cv=3)

            for score in scores:
                rank_scores.append(score)
                rank_values.append(rank_value)

        return pd.DataFrame({"rank": rank_values, "score": rank_scores})


def _rank_features(
    clf,
    X,
    y,
    scoring,
    replace,
    downsample_rate,
    rfe_step,
    rfe_min_features,
    rfe_importance_getter,
) -> tuple[RFE, dict]:
    """
    Rank features using Recursive Feature Elimination (RFE).

    Parameters:
    - clf: Classifier for feature selection.
    - X: Input data.
    - y: Target values.
    - scoring: Scoring metric for feature selection.
    - replace: Whether to use replacement during resampling.
    - downsample_rate: Rate of downsampling during resampling.
    - rfe_step: Step size for RFE.
    - rfe_min_features: Minimum number of features to select.
    - rfe_importance_getter: Method to get feature importances.

    Returns:
    Tuple: Tuple containing the feature selector and a dictionary of feature scores.
    """
    scores = dict()

    X_train, y_train = resample(
        X, y, replace=replace, n_samples=X.shape[0] * downsample_rate, stratify=y
    )

    selector = RFE(
        estimator=clf,
        step=rfe_step,
        n_features_to_select=rfe_min_features,
        importance_getter=rfe_importance_getter,
    )

    selector.fit(X_train, y_train)

    baseline_score = selector.score(X_train, y_train)

    print(f"Fit initial selector with score {baseline_score}.")

    for feature_num in np.unique(selector.ranking_):
        if isinstance(X_train, pd.DataFrame):
            X_subset = X_train.loc[:, (selector.ranking_ <= feature_num)]

        else:
            X_subset = X_train[:, (selector.ranking_ <= feature_num)]

        if X_subset.ndim < 2:
            X_subset = X_subset.reshape(-1, 1)

        score = cross_val_score(clf, X_subset, y_train, scoring=scoring, cv=3, n_jobs=1)

        scores[feature_num] = np.mean(score)

    return selector, scores



from typing import Any

import numpy as np
from sklearn.model_selection import cross_val_score, StratifiedKFold, RandomizedSearchCV
import random
from sklearn.base import clone


class ParamSearch:
    def __init__(self) -> None:
        pass

    def fit(self, classifier, X, y, **kwargs):
        pass


class ParamSearchResult:
    def __init__(self, classifier, result: Any) -> None:
        self.classifier = classifier
        self.result = result


class RandomizedSearch:
    def __init__(
        self,
        classifier,
        param_grid: dict,
        folds: int = 3,
        random_state: int = None,
        n_iter: int = 30,
        n_jobs: int = 4,
        scoring: str = "accuracy",
        verbose: int = 1,
    ):
        self.classifier = classifier
        self.param_grid = param_grid
        self.folds = folds
        self.random_state = random_state
        self.n_iter = n_iter
        self.n_jobs = n_jobs
        self.scoring = scoring
        self.verbose = verbose

    def fit(self, X, y) -> ParamSearchResult:
        skf = StratifiedKFold(
            n_splits=self.folds, shuffle=True, random_state=self.random_state
        )

        random_search = RandomizedSearchCV(
            self.classifier,
            param_distributions=self.param_grid,
            n_iter=self.n_iter,
            n_jobs=self.n_jobs,
            cv=skf.split(X, y),
            verbose=self.verbose,
            scoring=self.scoring,
            return_train_score=True,
        )

        random_search.fit(X, y)

        return ParamSearchResult(
            classifier=random_search.best_estimator_, result=random_search
        )


class GeneticAlgorithmSearch(ParamSearch):
    def __init__(
        self,
        classifier,
        param_grid: dict,
        n_generations: int = 50,
        pop_size: int = 10,
        n_survive: int = 5,
        threads: int = 1,
        folds: int = 3,
        verbose: bool = False,
        random_state: int = 42,
        shuffle: bool = False,
    ) -> None:
        self.classifier = classifier
        self.param_grid = param_grid
        self.threads = threads
        self.pop_size = pop_size
        self.folds = folds
        self.n_generations = n_generations
        self.n_survive = n_survive
        self.n_procreate = pop_size - n_survive
        self.populations = {}
        self.verbose = verbose
        self.random_state = random_state
        self.shuffle = shuffle

    def get_accuracy(self, classifier, X, y):
        skf = StratifiedKFold(
            n_splits=self.folds, shuffle=self.shuffle, random_state=self.random_state
        )
        scores = cross_val_score(classifier, X, y, n_jobs=self.threads, cv=skf)
        return np.mean(scores)

    def initiate_pop(self, param_grid):
        pop = []
        for _ in range(self.pop_size):
            random_initializer = random.uniform(0, 1)
            individual = {}
            for param in param_grid.keys():
                individual[param] = random.choice(param_grid[param])
            pop.append((random_initializer, individual))
        return pop

    def generation_pass(self, pop, X, y):
        evaluated_pop = []
        for _, params in pop:
            clf_individual = clone(self.classifier)
            clf_individual.set_params(**params)
            acc = self.get_accuracy(clf_individual, X, y)
            evaluated_pop.append((acc, params))
        return evaluated_pop

    def kill(self, pop):
        pop.sort(key=lambda x: x[0], reverse=True)
        reduced_pop = pop[: self.n_survive]
        return reduced_pop

    def procreate(self, pop):
        new_pop = pop.copy()
        for _ in range(self.n_procreate):
            individual = {}
            random_initializer = random.uniform(0, 1)
            for oldie in pop:
                _, old_individual = oldie
                for param in old_individual.keys():
                    individual[param] = random.choice(self.param_grid[param])
            new_pop.append((random_initializer, individual))
        return new_pop

    def fit(self, X, y) -> dict:
        pop = self.initiate_pop(self.param_grid)
        for generation in range(self.n_generations):
            evaluated_pop = self.generation_pass(pop, X, y)
            reduced_pop = self.kill(evaluated_pop)
            if self.verbose:
                print(f"Generation {generation}")
                print(f"Accuracy {reduced_pop[0][0]}")
                print(f"Best param {reduced_pop[0][1]}")
            self.populations[generation] = reduced_pop
            pop = self.procreate(reduced_pop)
        return self.populations

    @property
    def best_estimator_(self):
        if len(self.populations) == 0:
            raise Exception("The genetic algorithm has not been run")
        acc, best_params = self.populations[0][0]
        if self.verbose:
            print("Accuracy: ", acc)
        best_estimator = clone(self.classifier)
        best_estimator.set_params(**best_params)
        return best_estimator



#!/usr/bin/env python3
"""Tests for the quantification module"""

import pytest
from dpks.quant_matrix import QuantMatrix
from pathlib import Path
import pandas


@pytest.mark.quantification()
def test_quantification_protein(paths, quant_matrix):
    """quantification protein"""

    protein_quantities = quant_matrix.quantify(
        method="top_n",
        top_n=1,
    )

    assert type(protein_quantities) == QuantMatrix

    protein_quantities_df = protein_quantities.to_df()

    assert type(protein_quantities_df) == pandas.core.frame.DataFrame

    assert protein_quantities_df.shape == (570, 264)

    output_file = paths["test_base_path"] / Path("norm_prot_quant.tsv")
    protein_quantities_df.to_csv(output_file, sep="\t", index=False)

    assert output_file.is_file()



"""**scales quantitative matrices**, supports multiple methods as specified below

>>> from dpks.quant_matrix import QuantMatrix
>>> quant_matrix = QuantMatrix( quantification_file="tests/input_files/minimal_matrix.tsv", design_matrix_file="tests/input_files/minimal_design_matrix.tsv")
>>> quant_matrix.to_df()[["PeptideSequence", "SAMPLE_1.osw", "SAMPLE_2.osw", "SAMPLE_3.osw"]]
    PeptideSequence  SAMPLE_1.osw  SAMPLE_2.osw  SAMPLE_3.osw
0         EFMEEVIQR       69900.3      195571.0      403947.0
1  SSSGTPDLPVLLTDLK      115684.0      132524.0      217962.0
2            PEPTIK       29566.2       59295.7       24536.4

"""

from typing import TYPE_CHECKING, Any
import numpy as np

# import numpy.typing as npt  # not used yet

if TYPE_CHECKING:
    from .quant_matrix import QuantMatrix
else:
    QuantMatrix = Any


class ScalingMethod:
    """the base class"""

    def __init__(self) -> None:
        """init"""

        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """carry out the transform"""

        pass


class ZScoreScaling(ScalingMethod):
    """z-score scaling

    >>> from dpks.quant_matrix import QuantMatrix
    >>> quant_matrix = QuantMatrix( quantification_file="tests/input_files/minimal_matrix.tsv", design_matrix_file="tests/input_files/minimal_design_matrix.tsv")
    >>> zscore_scaled_df = quant_matrix.scale(method="zscore").to_df()[["PeptideSequence", "SAMPLE_1.osw", "SAMPLE_2.osw", "SAMPLE_3.osw"]]
    >>> zscore_scaled_df
        PeptideSequence  SAMPLE_1.osw  SAMPLE_2.osw  SAMPLE_3.osw
    0         EFMEEVIQR     -1.112361     -0.200119      1.312480
    1  SSSGTPDLPVLLTDLK     -0.886769     -0.510675      1.397444
    2            PEPTIK     -0.536779      1.401483     -0.864704
    >>> zscore_scaled_df.set_index("PeptideSequence").sum(axis=1)
    PeptideSequence
    EFMEEVIQR          -4.440892e-16
    SSSGTPDLPVLLTDLK    0.000000e+00
    PEPTIK              1.110223e-15
    dtype: float64

    """

    def __init__(self) -> None:
        """init"""

        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """carry out the z-score normalization"""

        means = np.mean(X, axis=1)
        stddevs = np.std(X, axis=1)

        return np.array((X - means[:, None]) / stddevs[:, None])


class MinMaxScaling(ScalingMethod):
    """min-max scaling

    Computes the min max scaling: (X - X_min) / (X_max - X_min)

    Resulting values are in range [0 1]
    """

    def __init__(self) -> None:
        """init"""

        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """carry out the min-max normalization"""

        mins = np.min(X, axis=1)
        maxes = np.max(X, axis=1)

        return np.array((X - mins[:, None]) / (maxes[:, None] - mins[:, None]))


class AbsMaxScaling(ScalingMethod):
    """abs-max scaling

    Computes the abs-max scaling: X / X_max

    Resulting values are in range [-1 1]
    """

    def __init__(self) -> None:
        """init"""

        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """carry out the min-max normalization"""

        maxes = np.max(X, axis=1)

        return np.array(X / maxes[:, None])



"""Console script for dpks."""
import sys
from typing import Union

import click


@click.command()
def main(args: Union[str, None] = None) -> int:
    """Console script for dpks."""
    click.echo("Replace this message by putting your code into " "dpks.cli.main")
    click.echo("See click documentation at https://click.palletsprojects.com/")
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python3
"""Testing the imputer module"""

import pytest
from dpks.quant_matrix import QuantMatrix
import pandas


@pytest.mark.imputer()
def test_imputer(quant_matrix):
    """test imputer"""

    assert isinstance(quant_matrix, QuantMatrix)

    imputed_peptide_quantities = quant_matrix.impute(
        method="uniform", minvalue=1, maxvalue=100
    )

    assert isinstance(imputed_peptide_quantities, QuantMatrix)

    imputed_peptide_quantities_df = imputed_peptide_quantities.to_df()

    assert isinstance(imputed_peptide_quantities_df, pandas.core.frame.DataFrame)



from typing import TYPE_CHECKING, Any

import numpy as np
from scipy import stats  # type: ignore

if TYPE_CHECKING:
    from .quant_matrix import QuantMatrix
else:
    QuantMatrix = Any

from statsmodels.stats.multitest import multipletests  # type: ignore


class DifferentialTest:
    method: str
    min_samples_per_group: int
    level: str
    group_a: int
    group_b: int
    multiple_testing_correction_method: str

    def __init__(
        self,
        method: str,
        comparisons: list,
        min_samples_per_group: int = 2,
        level: str = "precursor",
        multiple_testing_correction_method: str = "fdr_tsbh",
    ):
        self.method = method
        self.comparisons = comparisons
        self.min_samples_per_group = min_samples_per_group
        self.multiple_testing_correction_method = multiple_testing_correction_method
        if level == "precursor":
            self.level = "PrecursorId"

        elif level == "protein":
            self.level = "Protein"

        elif level == "peptide":
            self.level = "PeptideSequence"
        else:
            self.level = level

    def test(self, quant_matrix: QuantMatrix) -> QuantMatrix:
        if self.level == "PrecursorId":
            identifiers = quant_matrix.precursors

        elif self.level == "Protein":
            identifiers = quant_matrix.proteins

        elif self.level == "PeptideSequence":
            identifiers = quant_matrix.peptides
        else:
            identifiers = quant_matrix.quantitative_data.obs[self.level]

        if isinstance(self.comparisons, tuple):
            self.comparisons = [self.comparisons]

        for comparison in self.comparisons:
            group_a, group_b = comparison
            group_a_means = []
            group_b_means = []
            group_a_stdevs = []
            group_b_stdevs = []
            log_fold_changes = []
            p_values = []
            group_a_rep_counts = []
            group_b_rep_counts = []
            indices = []

            quant_matrix.quantitative_data.X[
                quant_matrix.quantitative_data.X == 0.0
            ] = np.nan

            for identifier in identifiers:
                quant_data = quant_matrix.quantitative_data[
                    quant_matrix.row_annotations[self.level] == identifier, :
                ].copy()

                indices.append(quant_data.obs.index.to_numpy()[0])

                group_a_samples = quant_matrix.get_samples(group=group_a)
                if self.method == "ttest_paired":
                    group_b_samples = quant_matrix.get_pairs(samples=group_a_samples)
                else:
                    group_b_samples = quant_matrix.get_samples(group=group_b)

                group_a_data = quant_data[:, group_a_samples].X.copy()
                group_b_data = quant_data[:, group_b_samples].X.copy()

                group_a_nan = len(group_a_data[~np.isnan(group_a_data)])
                group_b_nan = len(group_b_data[~np.isnan(group_b_data)])

                group_a_rep_counts.append(group_a_nan)
                group_b_rep_counts.append(group_b_nan)

                if (group_a_nan < self.min_samples_per_group) or (
                    group_b_nan < self.min_samples_per_group
                ):
                    if group_a_nan < self.min_samples_per_group:
                        group_a_means.append(np.nan)
                        group_a_stdevs.append(np.nan)

                    else:
                        group_a_means.append(np.mean(group_a_data))
                        group_a_stdevs.append(np.std(group_a_data))

                    if group_b_nan < self.min_samples_per_group:
                        group_b_means.append(np.nan)
                        group_b_stdevs.append(np.nan)

                    else:
                        group_b_means.append(np.mean(group_b_data))
                        group_b_stdevs.append(np.std(group_b_data))

                    log_fold_changes.append(np.nan)
                    p_values.append(np.nan)

                else:
                    group_a_data = group_a_data[~np.isnan(group_a_data)]
                    group_b_data = group_b_data[~np.isnan(group_b_data)]

                    group_a_mean = np.mean(group_a_data)
                    group_b_mean = np.mean(group_b_data)
                    group_a_stdev = np.std(group_a_data)
                    group_b_stdev = np.std(group_b_data)

                    log_fold_change = group_a_mean - group_b_mean

                    group_a_labels = np.array([1.0 for _ in range(len(group_a_data))])
                    group_b_labels = np.array([2.0 for _ in range(len(group_b_data))])

                    labels = np.concatenate([group_a_labels, group_b_labels])

                    expression_data = np.concatenate(
                        [group_a_data, group_b_data], axis=0
                    )

                    if self.method == "ttest":
                        test_results = stats.ttest_ind(group_a_data, group_b_data)

                    elif self.method == "linregress":
                        test_results = stats.linregress(x=expression_data, y=labels)

                    elif self.method == "anova":
                        test_results = stats.f_oneway(group_a_data, group_b_data)

                    elif self.method == "ttest_paired":
                        test_results = stats.ttest_rel(group_a_data, group_b_data)

                    group_a_means.append(group_a_mean)
                    group_b_means.append(group_b_mean)
                    group_a_stdevs.append(group_a_stdev)
                    group_b_stdevs.append(group_b_stdev)
                    log_fold_changes.append(log_fold_change)
                    p_values.append(test_results.pvalue)

            log_p_values = [-np.log(p) for p in p_values]
            max_log_p_value = np.nanmax(log_p_values)
            max_log_fold_change = np.nanmax([abs(fc) for fc in log_fold_changes])
            de_scores = [
                np.sqrt((p / max_log_p_value) ** 2 + (fc / max_log_fold_change) ** 2)
                for p, fc in zip(log_p_values, log_fold_changes)
            ]
            quant_matrix.row_annotations[f"DEScore{group_a}-{group_b}"] = de_scores
            quant_matrix.row_annotations[f"Group{group_a}Mean"] = group_a_means
            quant_matrix.row_annotations[f"Group{group_b}Mean"] = group_b_means
            quant_matrix.row_annotations[f"Group{group_a}Stdev"] = group_a_stdevs
            quant_matrix.row_annotations[f"Group{group_b}Stdev"] = group_b_stdevs
            quant_matrix.row_annotations[
                f"Log2FoldChange{group_a}-{group_b}"
            ] = log_fold_changes
            quant_matrix.row_annotations[f"PValues{group_a}-{group_b}"] = p_values
            quant_matrix.row_annotations[
                f"Group{group_a}RepCounts"
            ] = group_a_rep_counts
            quant_matrix.row_annotations[
                f"Group{group_b}RepCounts"
            ] = group_b_rep_counts

            quant_matrix.quantitative_data.obs.sort_values(
                f"PValues{group_a}-{group_b}", inplace=True
            )

            correction_results = multipletests(
                quant_matrix.quantitative_data.obs[
                    ~np.isnan(
                        quant_matrix.quantitative_data.obs[
                            f"PValues{group_a}-{group_b}"
                        ]
                    )
                ][f"PValues{group_a}-{group_b}"],
                method=self.multiple_testing_correction_method,
                is_sorted=False,
            )

            corrected_results = np.empty(
                (len(quant_matrix.quantitative_data.obs),), dtype=np.float64
            )
            corrected_results[:] = np.nan

            corrected_results[: len(correction_results[1])] = correction_results[1]

            quant_matrix.quantitative_data.obs[
                f"CorrectedPValue{group_a}-{group_b}"
            ] = corrected_results

            quant_matrix.quantitative_data.obs.index = (
                quant_matrix.quantitative_data.obs.index.map(int)
            )
            quant_matrix.quantitative_data.obs.sort_index(inplace=True)
            quant_matrix.quantitative_data.obs.index = (
                quant_matrix.quantitative_data.obs.index.map(str)
            )

        return quant_matrix



#!/usr/bin/env python3
"""Tests for the scaling module"""

import pytest
from dpks.quant_matrix import QuantMatrix
import pandas


@pytest.mark.scaling()
def test_scaling_zscore(quant_matrix):
    """zscore scaling"""

    assert isinstance(quant_matrix, QuantMatrix)

    scaled_peptide_quantities = quant_matrix.scale(method="zscore")

    assert isinstance(scaled_peptide_quantities, QuantMatrix)

    scaled_peptide_quantities_df = scaled_peptide_quantities.to_df()

    assert isinstance(scaled_peptide_quantities_df, pandas.core.frame.DataFrame)



from dpks.quant_matrix import QuantMatrix
import pytest
import xgboost


@pytest.fixture
def quantified_data(paths):
    qm = QuantMatrix(
        quantification_file=str(paths["sepsis_matrix_path"]),
        design_matrix_file=str(paths["sepsis_design_path"]),
    )
    quantified_data = qm.normalize(
        method="mean",
    ).quantify(method="top_n", summarization_method="mean")
    return quantified_data


def test_rfe(quantified_data: QuantMatrix):
    clf = xgboost.XGBClassifier()
    quantified_data.rank(clf, rfe_step=100, verbose=True, threads=4)



#!/usr/bin/env python3
"""Tests for the quant_matrix module"""

import pandas
from dpks.quant_matrix import QuantMatrix
import pytest


@pytest.mark.quant_matrix()
def test_quant_matrix_as_dataframe(quant_matrix):
    """test quant matrix as dataframe"""

    assert type(quant_matrix) == QuantMatrix

    quant_matrix_df = quant_matrix.to_df()

    assert type(quant_matrix_df) == pandas.core.frame.DataFrame



import seaborn as sns  # type: ignore # noqa: F401
from typing import TYPE_CHECKING, Any, Union, List, Tuple, Optional

import numpy as np
import pandas as pd
import matplotlib
import sklearn

import matplotlib.pyplot as plt
from matplotlib import cm


if TYPE_CHECKING:
    from .quant_matrix import QuantMatrix
else:
    QuantMatrix = Any


class Plot:
    """the base class"""

    def __init__(self) -> None:
        """init"""

        pass

    def plot(self) -> None:
        """create the plot"""
        pass


class SHAPPlot(Plot):
    def __init__(
        self,
        fig: matplotlib.figure.Figure,
        ax: matplotlib.axes.Axes,
        shap_values: np.ndarray,
        X: np.ndarray,
        qm: QuantMatrix,
        cmap: Union[List, str],
        n_display: int = 5,
        jitter: float = 0.1,
        alpha: float = 0.75,
        feature_column: str = "Protein",
        order_by: str = "shap",
        n_bins=100,
    ):
        """Creates a SHAP summary plot-like figure.

        Args:
            shap_values (np.ndarray): shap values
            X (np.ndarray): feature values
            qm (QuantMatrix): quantmatrix

        Returns:
            plt.Figure: figure object
        """
        self.shap_values = shap_values
        self.order_by = order_by
        self.X = X
        self.qm = qm
        self.n_display = n_display
        self.cmap = cmap
        self.jitter = jitter
        self.alpha = alpha
        self.feature_column = feature_column
        self.n_bins = n_bins

        if not fig:
            self.fig, self.ax = plt.subplots(
                1, 1, figsize=(n_display * 1.5, n_display * 1)
            )
        else:
            self.fig = fig
            self.ax = ax

    def plot(self) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        plot_frame = pd.DataFrame(columns=["feature", "x", "y"])
        col_sum = np.mean(np.abs(self.shap_values), axis=0)

        if self.order_by == "shap":
            sort_index = np.argsort(-col_sum)
        elif self.order_by == "rank":
            sort_index = np.argsort(self.qm.row_annotations["FeatureRank"])

        feature_names = []
        for idx, feature_idx in enumerate(sort_index[0 : self.n_display]):
            feature_name = self.qm.quantitative_data.obs[self.feature_column][
                feature_idx
            ]
            feature_names.append(feature_name)
            sv = self.shap_values[:, feature_idx]
            fv = self.X[:, feature_idx]
            plot_frame = pd.concat(
                [
                    plot_frame,
                    pd.DataFrame(
                        data={
                            "feature": len(fv) * [feature_name],
                            "sv": sv,
                            "fv": fv,
                            "feature_idx": idx,
                        }
                    ),
                ]
            )

        if isinstance(self.cmap, list):
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
                "customcmap", self.cmap
            )
        else:
            cmap = plt.get_cmap(self.cmap)

        v = min(np.abs(plot_frame["fv"].min()), plot_frame["fv"].max())
        norm = matplotlib.colors.Normalize(vmin=-v, vmax=v)

        colors = {}
        for cval in plot_frame["fv"]:
            colors.update({cval: cmap(norm(cval))})

        self.ax.axvline([0], c="#f5f5f5", zorder=-1)

        plot_frame = self._jitter(plot_frame)
        plot_frame["y"] = plot_frame["feature_idx"] + plot_frame["jitter"]
        c = plot_frame["fv"].map(colors)
        sns.violinplot(
            data=plot_frame,
            x="sv",
            y="feature",
            alpha=0.1,
            color="lightgray",
            linewidth=0,
            ax=self.ax,
            scale="count",
        )
        sns.scatterplot(
            x=plot_frame["sv"],
            y=plot_frame["y"],
            c=c,
            alpha=self.alpha,
            linewidth=0,
            ax=self.ax,
        )
        cb = self.fig.colorbar(cm.ScalarMappable(cmap=cmap, norm=norm), ax=self.ax)

        cb.outline.set_visible(False)
        sns.despine(ax=self.ax, left=True, right=True, top=True)

        self.ax.set_xlabel("SHAP value")
        self.ax.set_ylabel("Feature")
        self.ax.set_yticks(range(self.n_display), feature_names)
        return self.fig, self.ax

    def _jitter(self, plot_frame):
        plot_frame["bin"] = pd.cut(
            plot_frame["sv"], bins=self.n_bins, labels=range(self.n_bins)
        )
        bins = plot_frame["bin"].value_counts()

        bins_desc = bins.index.values.tolist()
        bins_desc.reverse()
        jitters = []
        for row in plot_frame.iterrows():
            b = row[1].bin
            i = bins_desc.index(b)
            jitter = np.random.normal(scale=self.jitter) * i / self.n_bins
            jitters.append(jitter)
        plot_frame["jitter"] = jitters
        return plot_frame


class RFEPCA(Plot):
    def __init__(
        self,
        fig: matplotlib.figure.Figure,
        axs: List[matplotlib.axes.Axes],
        qm: QuantMatrix,
        cutoffs: List,
        cmap: Union[list, str],
    ):
        self.qm = qm
        self.cutoffs = cutoffs
        self.fig = fig
        if isinstance(axs, list):
            axs = np.array(axs)
        self.axs = axs
        self.cmap = cmap
        if not fig:
            self.fig, self.axs = plt.subplots(1, 1, figsize=(5, 5))

    def plot(self) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        qdf = self.qm.to_df()

        samples1 = self.qm.get_samples(group=1)
        samples2 = self.qm.get_samples(group=2)

        y = [1 for _ in samples1] + [2 for _ in samples2]
        if isinstance(self.cmap, list):
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
                "customcmap", self.cmap
            )
        else:
            cmap = plt.get_cmap(self.cmap)
        for ax, cutoff in zip(self.axs.ravel(), self.cutoffs):
            qdf_f = qdf[qdf.FeatureRank <= cutoff]
            qdf_f = qdf_f[samples1 + samples2]
            X = qdf_f.values.T
            X = sklearn.preprocessing.StandardScaler().fit_transform(X)
            pca = sklearn.decomposition.PCA(n_components=2)
            pca.fit(X)
            X = pca.transform(X)
            explained_variance = pca.explained_variance_ratio_

            sns.kdeplot(
                x=X[:, 0],
                y=X[:, 1],
                levels=5,
                hue=y,
                palette=cmap,
                ax=ax,
                fill=True,
                alpha=0.2,
            )

            sns.kdeplot(
                x=X[:, 0],
                y=X[:, 1],
                levels=5,
                hue=y,
                palette=cmap,
                linewidths=2,
                ax=ax,
            )

            sns.scatterplot(x=X[:, 0], y=X[:, 1], hue=y, palette=cmap, ax=ax)

            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xlabel(f"PC1 ({100*explained_variance[0]:.1f}%)")
            ax.set_ylabel(f"PC2 ({100*explained_variance[1]:.1f}%)")

        return self.fig, self.axs



#!/usr/bin/env python3
"""Tests for the normalization module"""

import pytest
from dpks.quant_matrix import QuantMatrix


@pytest.mark.normalization()
def test_normalization_mean(quant_matrix):
    """normalization mean"""

    assert isinstance(quant_matrix, QuantMatrix)

    normalized_peptide_quantities = quant_matrix.normalize(method="mean")

    assert isinstance(normalized_peptide_quantities, QuantMatrix)


@pytest.mark.normalization()
def test_normalization_tic(quant_matrix):
    """normalization tic"""

    assert isinstance(quant_matrix, QuantMatrix)

    normalized_peptide_quantities = quant_matrix.normalize(method="tic")

    assert isinstance(normalized_peptide_quantities, QuantMatrix)


@pytest.mark.normalization()
def test_normalization_median(quant_matrix):
    """normalization median"""

    assert isinstance(quant_matrix, QuantMatrix)

    normalized_peptide_quantities = quant_matrix.normalize(method="median")

    assert isinstance(normalized_peptide_quantities, QuantMatrix)


@pytest.mark.normalization()
def test_normalization_rt_median(quant_matrix):
    """normalization median"""

    assert isinstance(quant_matrix, QuantMatrix)

    normalized_peptide_quantities = quant_matrix.normalize(
        method="mean",
        log_transform=True,
        use_rt_sliding_window_filter=True,
        minimum_data_points=100,
        stride=5,
        use_overlapping_windows=True,
        rt_unit="seconds",
    )

    assert isinstance(normalized_peptide_quantities, QuantMatrix)



"""**normalizes quantitative matrices**, supports multiple methods as specified below"""
import warnings

from typing import TYPE_CHECKING, Any, List

import numpy as np
import pandas as pd  # type: ignore

if TYPE_CHECKING:
    from .quant_matrix import QuantMatrix
else:
    QuantMatrix = Any


class NormalizationMethod:
    def __init__(self) -> None:
        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        pass


class Log2Normalization(NormalizationMethod):
    def __init__(self) -> None:
        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        return np.array(np.log2(X), dtype=np.float64)


class TicNormalization(NormalizationMethod):
    def __init__(self) -> None:
        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:

        with warnings.catch_warnings():

            warnings.simplefilter("ignore", category=RuntimeWarning)

            sample_sums = np.nansum(X, axis=0)

        median_signal = np.nanmedian(sample_sums)

        normalized_signal: np.ndarray = (X / sample_sums[None, :]) * median_signal

        return normalized_signal


class MedianNormalization(NormalizationMethod):
    def __init__(self) -> None:
        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:

        with warnings.catch_warnings():

            warnings.simplefilter("ignore", category=RuntimeWarning)

            sample_medians = np.nanmedian(X, axis=0)

        mean_sample_median = np.mean(sample_medians)

        normalized_signal: np.ndarray = (
            X / sample_medians[None, :]
        ) * mean_sample_median

        return normalized_signal


class MeanNormalization(NormalizationMethod):
    """normalize the quantitative data using **mean**

    The input data is found in the DPKS git reposotory:

    - The minimal design matrix: `minimal_design_matrix.tsv`_. -  contains the experimental design
    - The minimal matrix: `minimal_matrix.tsv`_.  - contains the quantitative data

    .. _minimal_design_matrix.tsv: https://github.com/InfectionMedicineProteomics/DPKS/blob/doctest001/tests/input_files/minimal_design_matrix.tsv
    .. _minimal_matrix.tsv: https://github.com/InfectionMedicineProteomics/DPKS/blob/doctest001/tests/input_files/minimal_matrix.tsv

    The QuantMatrix instance is created by providing a quantitative matrix and a design matrix:

    >>> from dpks.quant_matrix import QuantMatrix
    >>> quant_matrix = QuantMatrix( quantification_file="tests/input_files/minimal_matrix.tsv", design_matrix_file="tests/input_files/minimal_design_matrix.tsv")
    >>> quant_matrix.to_df().sort_values(by="PeptideSequence").set_index("PeptideSequence").head(1).T
    PeptideSequence EFMEEVIQR
    Charge                  2
    Decoy                   0
    Protein            P04275
    RetentionTime      3155.5
    PeptideQValue    0.000009
    ProteinQValue    0.000117
    SAMPLE_1.osw      69900.3
    SAMPLE_2.osw     195571.0
    SAMPLE_3.osw     403947.0

    Applying the normalize method and specifying the method:

    >>> norm_quant_matrix = quant_matrix.normalize(method="mean")
    >>> norm_quant_matrix.to_df()[["PeptideSequence", "SAMPLE_1.osw", "SAMPLE_2.osw", "SAMPLE_3.osw"]].sort_values(by="PeptideSequence").set_index("PeptideSequence").head(1).T
    PeptideSequence  EFMEEVIQR
    SAMPLE_1.osw     17.045388
    SAMPLE_2.osw     17.681267
    SAMPLE_3.osw     17.989002

    """

    #        PeptideSequence  SAMPLE_1.osw  SAMPLE_2.osw  SAMPLE_3.osw
    #    0         EFMEEVIQR     17.045388     17.681267     17.989002
    #    2            PEPTIK     15.804039     15.959574     13.947831
    #    1  SSSGTPDLPVLLTDLK     17.772207     17.119828     17.098912

    def __init__(self) -> None:
        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:

        with warnings.catch_warnings():

            warnings.simplefilter("ignore", category=RuntimeWarning)

            sample_means = np.nanmean(X, axis=0)

        mean_sample_means = np.mean(sample_means)

        normalized_signal: np.ndarray = (X / sample_means[None, :]) * mean_sample_means

        return normalized_signal


class RTSlidingWindowNormalization:
    base_method: NormalizationMethod
    minimum_data_points: int
    stride: float
    use_overlapping_windows: bool
    rt_unit: str

    def __init__(
        self,
        base_method: NormalizationMethod,
        stride: float = 1.0,
        minimum_data_points: int = 250,
        use_overlapping_windows: bool = False,
        rt_unit: str = "minute",
    ):
        self.base_method = base_method
        self.stride = stride
        self.minimum_data_points = minimum_data_points
        self.use_overlapping_windows = use_overlapping_windows
        self.rt_unit = rt_unit

    def build_rt_windows(self, rts: np.ndarray) -> List[np.ndarray]:
        if self.rt_unit == "seconds":
            rts = rts / 60.0

        rt_min = np.min(rts)
        rt_max = np.max(rts)

        rt_bins = np.arange(rt_min, rt_max, self.stride)

        rt_indices = np.digitize(rts, rt_bins) - 1

        bin_indices = []

        for rt_idx in range(len(rt_bins)):
            indices = np.argwhere(rt_indices == rt_idx)

            if indices.reshape(-1).shape[0] < self.minimum_data_points:
                expanded_indices = []

                expanded_indices.append(indices)

                for expand_index in range(1, len(rt_bins)):
                    lower_bound_idx = rt_idx - expand_index

                    upper_bound_idx = rt_idx + expand_index

                    if lower_bound_idx < 0:
                        expanded_indices.append(
                            np.argwhere(rt_indices == upper_bound_idx)
                        )

                    elif expand_index >= len(rt_bins):
                        expanded_indices.append(
                            np.argwhere(rt_indices == lower_bound_idx)
                        )

                    else:
                        expanded_indices.append(
                            np.argwhere(rt_indices == lower_bound_idx)
                        )

                        expanded_indices.append(
                            np.argwhere(rt_indices == upper_bound_idx)
                        )

                    if (
                        np.concatenate(expanded_indices).reshape(-1).shape[0]
                        >= self.minimum_data_points
                    ):
                        break

                bin_indices.append(np.concatenate(expanded_indices).reshape(-1))

            else:
                bin_indices.append(indices.reshape(-1))

        return bin_indices

    def overlap_rt_windows(self, rt_windows: List[np.ndarray]) -> List[np.ndarray]:
        new_rt_windows = []

        for rt_idx, rt_window in enumerate(rt_windows):
            if rt_idx == 0:
                new_rt_window = np.concatenate(
                    [
                        rt_window,
                        rt_windows[rt_idx + 1],
                    ]
                )

            elif rt_idx == len(rt_windows) - 1:
                new_rt_window = np.concatenate(
                    [
                        rt_window,
                        rt_windows[rt_idx - 1],
                    ]
                )

            else:
                new_rt_window = np.concatenate(
                    [
                        rt_windows[rt_idx - 1],
                        rt_window,
                        rt_windows[rt_idx + 1],
                    ]
                )

            new_rt_windows.append(new_rt_window)

        return new_rt_windows

    def fit_transform(self, quantitative_data: QuantMatrix) -> np.ndarray:
        normalized_slices = []

        rt_windows = self.build_rt_windows(
            rts=quantitative_data.quantitative_data.obs["RetentionTime"]
        )

        if self.use_overlapping_windows:
            rt_windows = self.overlap_rt_windows(rt_windows=rt_windows)

        for rt_window in rt_windows:
            normalized_slice = pd.DataFrame(
                self.base_method.fit_transform(
                    quantitative_data.quantitative_data[rt_window, :].X.copy()
                )
            ).set_index(pd.Index(rt_window))

            normalized_slices.append(normalized_slice)

        joined_slices = pd.concat(normalized_slices)

        joined_slices = joined_slices.reset_index().groupby("index", sort=False).mean()

        return_array: np.ndarray = joined_slices.to_numpy(dtype=np.float64)

        return return_array



from typing import Optional, List

import pandas as pd

from sklearn.utils import resample

from dpks.classification import Classifier

from imblearn.under_sampling import RandomUnderSampler
from kneed import KneeLocator


class BootstrapInterpreter:
    def __init__(
        self,
        n_iterations: int = 10,
        feature_names: Optional[List[str]] = None,
        downsample_background: bool = False,
    ):
        self.percent_cutoff = None
        self.feature_counts = None
        self.n_iterations = n_iterations
        self.feature_names = feature_names
        self.downsample_background = downsample_background

    def fit(self, X, y, classifier) -> None:
        results = dict()

        results["feature"] = self.feature_names

        for i in range(self.n_iterations):
            X_train, y_train = resample(
                X, y, replace=True, n_samples=X.shape[0] * 1, stratify=y, random_state=i
            )

            if isinstance(classifier, Classifier):
                clf = classifier
            else:
                clf = Classifier(classifier=classifier)

            clf.fit(X_train, y_train)

            if self.downsample_background:
                rus = RandomUnderSampler(random_state=0)
                X_resampled, y_resampled = rus.fit_resample(X_train, y_train)
                clf.interpret(X_resampled)
            else:
                clf.interpret(X_train)

            results[f"iteration_{i}_shap"] = pd.Series(
                clf.mean_importance / clf.mean_importance.max()
            )
            results[f"iteration_{i}_rank"] = results[f"iteration_{i}_shap"].rank(
                ascending=False
            )

        self.importances = pd.DataFrame(results)

        self.importances["mean_shap"] = self.importances[
            [f"iteration_{i}_shap" for i in range(self.n_iterations)]
        ].mean(axis=1)
        self.importances["median_shap"] = self.importances[
            [f"iteration_{i}_shap" for i in range(self.n_iterations)]
        ].median(axis=1)
        self.importances["stdev_shap"] = self.importances[
            [f"iteration_{i}_shap" for i in range(self.n_iterations)]
        ].std(axis=1)

        self.importances["mean_rank"] = self.importances[
            [f"iteration_{i}_rank" for i in range(self.n_iterations)]
        ].mean(axis=1)
        self.importances["median_rank"] = self.importances[
            [f"iteration_{i}_rank" for i in range(self.n_iterations)]
        ].median(axis=1)
        self.importances["stdev_rank"] = self.importances[
            [f"iteration_{i}_rank" for i in range(self.n_iterations)]
        ].std(axis=1)

    @property
    def results_(self) -> pd.DataFrame:
        return self.importances

    def select_features(
        self,
        top_n: int = 10,
        percent: float = 0.5,
        method: str = "shap",
        metric="percent",
    ) -> List[str]:
        final_features = list()
        all_selected_features = dict()

        if method == "count":
            for i in range(self.n_iterations):
                selected_features = (
                    self.importances.sort_values(f"iteration_{i}_shap", ascending=False)
                    .head(top_n)["feature"]
                    .to_list()
                )

                for feature in selected_features:
                    all_selected_features[feature] = (
                        all_selected_features.get(feature, 0) + 1
                    )

            feature_counts = {
                k: v / self.n_iterations
                for k, v in sorted(
                    all_selected_features.items(),
                    key=lambda item: item[1],
                    reverse=True,
                )
            }

            self.feature_counts = pd.DataFrame(
                {"feature": feature_counts.keys(), "count": feature_counts.values()}
            )

            if metric == "percent":
                self.percent_cutoff = percent

                final_features = self.feature_counts[
                    self.feature_counts["count"] > percent
                ]["feature"].to_list()

            elif metric == "knee":
                sorted_feature_counts = self.feature_counts.sort_values(
                    "count", ascending=False
                ).reset_index(drop=True)

                kn = KneeLocator(
                    sorted_feature_counts.index.values,
                    sorted_feature_counts["count"].values,
                    curve="convex",
                    direction="decreasing",
                )

                self.percent_cutoff = kn.knee_y

                final_features = self.feature_counts[
                    self.feature_counts["count"] > self.percent_cutoff
                ]["feature"].to_list()

        elif method == "shap":
            # self.importances['shap_scaled'] = self.importances['mean_shap'] / self.importances['mean_shap'].sum()

            if metric == "percent":
                self.percent_cutoff = percent

                final_features = self.importances[
                    self.importances["mean_shap"] > self.percent_cutoff
                ]["feature"].to_list()

            elif metric == "knee":
                sorted_feature_counts = self.importances.sort_values(
                    "mean_shap", ascending=False
                ).reset_index(drop=True)

                kn = KneeLocator(
                    sorted_feature_counts.index.values,
                    sorted_feature_counts["mean_shap"].values,
                    curve="convex",
                    direction="decreasing",
                )

                self.percent_cutoff = kn.knee_y

                final_features = self.importances[
                    self.importances["mean_shap"] > self.percent_cutoff
                ]["feature"].to_list()

        return final_features



import warnings
from typing import List, Any, TYPE_CHECKING, Tuple, Union

# from itertools import combinations, repeat  # not used yet
# from multiprocessing import Pool  # not used yet

#  import warnings  # not used yet

import numba  # type: ignore
import numpy as np

# import numpy.typing as npt  # not used yet
import pandas as pd  # type: ignore
from numba import njit, prange  # type: ignore

# from numpy.linalg import LinAlgError  # type: ignore

if TYPE_CHECKING:
    from .quant_matrix import QuantMatrix
else:
    QuantMatrix = Any


class TopN:
    top_n: int

    def __init__(
        self, top_n: int = 1, level: str = "protein", summarization_method: str = "sum"
    ):
        self.top_n = top_n
        self.num_proteins = 0
        self.num_samples = 0

        self.level = level
        self.summarization_method = summarization_method

    def quantify(self, quant_matrix: QuantMatrix) -> pd.DataFrame:
        quantifications = dict()

        if self.level == "protein":
            key = "Protein"
            group_ids = quant_matrix.proteins

        elif self.level == "precursor":
            key = "PrecursorId"
            group_ids = quant_matrix.precursors

        elif self.level == "peptide":
            key = "PeptideSequence"
            group_ids = quant_matrix.peptides

        elif self.level in quant_matrix.quantitative_data.obs.columns:
            key = self.level
            group_ids = list(quant_matrix.quantitative_data.obs[key].unique())
        else:
            raise ValueError("The level is not valid.")

        for group_id in group_ids:
            group_data = quant_matrix.quantitative_data[
                quant_matrix.quantitative_data.obs[key] == group_id
            ]

            quantification = self.quantify_group(group_data.X)

            quantifications[group_id] = quantification

        results = pd.DataFrame(quantifications)

        results = results.T.copy()

        results.columns = list(quant_matrix.quantitative_data.var["sample"])

        return results.reset_index().rename(columns={"index": key})

    def quantify_group(self, group_data: np.ndarray) -> np.ndarray:
        # Need to temporarily convery this so that the data can be sorted correctly
        group_data = np.nan_to_num(group_data, nan=0.0)

        sort_indices = np.argsort(group_data, axis=0)[::-1]

        # Covert back to array since ArrayView cannot be boolean indexed to change 0 back to np.NaN
        sorted_precursors = np.array(
            np.take_along_axis(group_data, sort_indices, axis=0)
        )

        sorted_precursors[sorted_precursors == 0.0] = np.nan

        if self.summarization_method == "sum":
            quantification: np.ndarray = np.nansum(
                sorted_precursors[: self.top_n], axis=0
            )

        elif self.summarization_method == "mean":
            quantification: np.ndarray = np.nanmean(
                sorted_precursors[: self.top_n], axis=0
            )

        elif self.summarization_method == "median":
            quantification: np.ndarray = np.nanmedian(
                sorted_precursors[: self.top_n], axis=0
            )

        return quantification


@njit(nogil=True, nopython=True)
def get_ratios(quantitative_data, sample_combinations):
    num_samples = quantitative_data.shape[1]

    ratios = np.empty((num_samples, num_samples), dtype=np.float64)

    ratios[:] = np.nan

    for combination in sample_combinations:
        sample_a = combination[0]
        sample_b = combination[1]

        ratio = -quantitative_data[:, sample_a] + quantitative_data[:, sample_b]

        ratio_median = np.nanmedian(ratio)

        ratios[sample_b, sample_a] = ratio_median

    return ratios


@njit(nogil=True, nopython=True)
def solve_profile(X, ratios, sample_combinations):
    if np.all(np.isnan(X)):
        results = np.zeros((X.shape[1]))

    else:
        num_samples = X.shape[1]

        A = np.zeros((num_samples + 1, num_samples + 1))
        b = np.zeros((num_samples + 1,))

        for sample_combination in sample_combinations:
            i = sample_combination[0]
            j = sample_combination[1]

            A[i][j] = -1.0
            A[j][i] = -1.0
            A[i][i] += 1.0
            A[j][j] += 1.0

            ratio = ratios[j, i]

            if not np.isnan(ratio):
                b[i] -= ratio
                b[j] += ratio

        formatted_a = 2.0 * A
        formatted_a[:num_samples, num_samples] = 1
        formatted_a[num_samples, :num_samples] = 1

        formatted_b = 2.0 * b

        sample_mean = np.nanmean(X)

        if np.isnan(sample_mean):
            sample_mean = 0.0

        formatted_b[num_samples] = sample_mean * num_samples

        nan_idx = np.argwhere(np.isnan(b))

        for nan_value in nan_idx:
            formatted_b[nan_value] = 0.0

        results = np.linalg.lstsq(formatted_a, formatted_b, -1.0)[0][: X.shape[1]]

    results[results == 0.0] = np.nan

    return results


@njit(nogil=True, nopython=True)
def build_connection_graph(grouping):
    connected_sample_groups = numba.typed.Dict()

    connected_indices = numba.typed.List()

    sample_group_id = 0

    for sample_idx in range(grouping.shape[1]):
        if sample_idx not in connected_indices:
            sample_group = []

            for compared_sample_idx in range(grouping.shape[1]):
                comparison = grouping[:, sample_idx] - grouping[:, compared_sample_idx]

                if not np.isnan(comparison).all():
                    sample_group.append(compared_sample_idx)

                    connected_indices.append(compared_sample_idx)

            if len(sample_group) > 0:
                connected_sample_groups[sample_group_id] = np.array(sample_group)

                sample_group_id += 1

            else:
                connected_sample_groups[sample_group_id] = np.array([sample_idx])

                sample_group_id += 1

    return connected_sample_groups


@njit(nogil=True, nopython=True)
def build_combinations(subset):
    column_idx = np.arange(0, subset.shape[1])

    combos = []

    for i in column_idx:
        for j in range(i + 1, column_idx.shape[0]):
            combos.append([i, j])

    return np.array(combos)


@njit(nogil=True, nopython=True)
def mask_group(grouping):
    nan_groups = []

    for subgroup_idx in range(grouping.shape[0]):
        if not np.isnan(grouping[subgroup_idx, :]).all():
            nan_groups.append(subgroup_idx)

    grouping = grouping[np.array(nan_groups), :]

    return grouping


@njit(nogil=True, nopython=True)
def quantify_group(grouping, connected_graph):
    profile = np.zeros((grouping.shape[1]))

    for sample_group_id, graph in connected_graph.items():
        if graph.shape[0] == 1:
            subset = grouping[:, graph]

            if np.isnan(subset).all():
                profile[graph] = np.nan

            else:
                profile[graph] = np.nanmedian(subset)

        if graph.shape[0] > 1:
            subset = grouping[:, graph]

            sample_combinations = build_combinations(subset)

            ratios = get_ratios(subset, sample_combinations)

            solved_profile = solve_profile(subset, ratios, sample_combinations)

            for results_idx in range(solved_profile.shape[0]):
                profile[graph[results_idx]] = solved_profile[results_idx]

    return profile


@njit(parallel=True, nopython=True)
def quantify_groups(groupings, group_ids, minimum_subgroups):
    num_groups = len(group_ids)

    results = np.empty(shape=(num_groups, groupings[0].shape[1]))

    result_ids = ["" for _ in range(num_groups)]

    for group_idx in prange(num_groups):
        grouping = mask_group(groupings[group_idx])

        if grouping.shape[0] >= minimum_subgroups:
            connected_graph = build_connection_graph(grouping)

            profile = quantify_group(grouping, connected_graph)

        else:
            profile = np.zeros((grouping.shape[1]))
            profile[:] = np.nan

        for sample_idx in range(profile.shape[0]):
            results[group_idx, sample_idx] = profile[sample_idx]

        result_ids[group_idx] = group_ids[group_idx]

    return_results = numba.typed.List()

    for group_idx in range(num_groups):
        return_results.append((result_ids[group_idx], results[group_idx, :]))

    return return_results


class MaxLFQ:
    level: str
    threads: int
    minimum_subgroups: int
    top_n: int

    def __init__(
        self,
        level: str = "protein",
        threads: int = 1,
        minimum_subgroups: int = 1,
        top_n: int = 0,
    ):
        self.level = level
        self.threads = threads
        self.minimum_subgroups = minimum_subgroups
        self.top_n = top_n

        numba.set_num_threads(threads)

    def quantify(self, quant_matrix: QuantMatrix) -> pd.DataFrame:
        group_ids = numba.typed.List()
        key: str = ""

        if self.level == "protein":
            group_ids = numba.typed.List(quant_matrix.proteins)
            key = "Protein"

        elif self.level == "precursor":
            group_ids = numba.typed.List(quant_matrix.precursors)
            key = "PrecursorId"

        elif self.level == "peptide":
            group_ids = numba.typed.List(quant_matrix.peptides)
            key = "PeptideSequence"

        elif self.level in quant_matrix.quantitative_data.obs.columns:
            group_ids = numba.typed.List(
                quant_matrix.quantitative_data.obs[self.level].unique()
            )
            key = self.level
        else:
            raise ValueError("The level is not valid.")

        groupings = numba.typed.List()

        for group_id in group_ids:
            group_data = quant_matrix.quantitative_data[
                quant_matrix.quantitative_data.obs["Protein"] == group_id
            ].copy()

            with warnings.catch_warnings():

                warnings.filterwarnings('error')

                try:

                    group_data.obs["MeanAbundance"] = np.nanmean(group_data.X, axis=1)

                except RuntimeWarning:

                    group_data.obs["MeanAbundance"] = 0

            sort_indices = np.argsort(group_data.obs["MeanAbundance"].values)[::-1]

            sorted_precursors = np.array(
                np.take_along_axis(group_data.X, sort_indices.reshape((-1, 1)), axis=0)
            )

            if self.top_n > 0:
                sorted_precursors = sorted_precursors[: self.top_n]

            groupings.append(sorted_precursors)

        quantified_groups = quantify_groups(
            groupings, group_ids, self.minimum_subgroups
        )

        if self.level == "protein":
            results = self._format_result_df(
                quantified_groups, quant_matrix.sample_annotations["sample"].values
            )

        elif self.level == "precursor":
            results = self._format_result_df(
                quantified_groups, quant_matrix.sample_annotations["sample"].values
            )

            results = results.set_index(key).join(
                quant_matrix.quantitative_data.obs.set_index(key)
            )

        elif self.level == "peptide":
            results = self._format_result_df(
                quantified_groups, quant_matrix.sample_annotations["sample"].values
            )

            results = (
                results.set_index(key)
                .merge(
                    quant_matrix.quantitative_data.obs[
                        ~quant_matrix.quantitative_data.obs.duplicated(
                            ["PeptideSequence", "Protein"],
                        )
                    ].set_index(key),
                    how="left",
                    left_index=True,
                    right_index=True,
                )
                .reset_index()
            )
        else:
            results = self._format_result_df(
                quantified_groups, quant_matrix.sample_annotations["sample"].values
            )

        return results

    def _format_result_df(
        self,
        quantified_groups: List[Tuple[str, Union[np.ndarray, np.ndarray], bool]],
        samples: np.ndarray,
    ) -> pd.DataFrame:
        rows = []

        if self.level == "protein":
            label = "Protein"

        elif self.level == "precursor":
            label = "PrecursorId"

        elif self.level == "peptide":
            label = "PeptideSequence"

        else:
            label = self.level

        for result in quantified_groups:
            result_id = result[0]
            result_data = result[1]

            row = {
                label: result_id,
            }

            for idx, sample in enumerate(samples):
                row[sample] = result_data[idx]

            rows.append(row)

        return pd.DataFrame(rows)



from typing import List

from Bio import SeqIO
import gzip


def parse_fasta(fasta):
    fasta_dict = {}
    for record in SeqIO.parse(fasta, "fasta"):
        id = record.id

        if "|" in id:
            accession = id.split("|")[1]
            name = id.split("|")[2]
        else:
            accession = id
            name = id
        fasta_dict[accession] = name
    return fasta_dict


def fasta_to_dict(fasta_path: str) -> dict:
    fasta_dict = {}
    if fasta_path.endswith("gz"):
        with gzip.open(fasta_path, "rt") as handle:
            fasta_dict = parse_fasta(handle)
    else:
        fasta_dict = parse_fasta(fasta_path)
    return fasta_dict


def get_protein_labels(accessions: List[str], fasta_path: str) -> List[str]:
    """Maps UniProt ID (accession) to more common names. P69905 --> HBA_HUMAN
    Args:
        accessions list[str]: Input protein list
        fasta_path str: fasta file path used for mapping
    Returns:
        protein_labels list[str]: More common protein names. If accession not in human_proteome.gz, use accession.
    """
    fasta_dict = fasta_to_dict(fasta_path)
    protein_labels = []
    for accession in accessions:
        try:
            protein_label = fasta_dict[accession]
        except KeyError:
            # If accession not in fasta file, put accession instead of protein label
            protein_label = accession
        protein_labels.append(protein_label)
    return protein_labels



from typing import TYPE_CHECKING, Any
import xgboost
from sklearn.model_selection import cross_val_score
import numpy as np
import shap
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import LabelEncoder
from sklearn.utils import class_weight

if TYPE_CHECKING:
    from .quant_matrix import QuantMatrix
else:
    QuantMatrix = Any


class TrainResult:
    """
    Class to store the results of a training process.

    Parameters:
    - classifier: Trained classifier.
    - scaler: Scaler used during training.
    - validation_results: Results from the validation process.

    Attributes:
    - classifier: Trained classifier.
    - scaler: Scaler used during training.
    - validation_results: Results from the validation process.
    - estimator_: Alias for the trained classifier.

    """

    def __init__(self, classifier, scaler, validation_results):
        """
        Initialize TrainResult with classifier, scaler, and validation results.

        Parameters:
        - classifier: Trained classifier.
        - scaler: Scaler used during training.
        - validation_results: Results from the validation process.

        Returns:
        None
        """
        self.classifier = classifier
        self.scaler = scaler
        self.validation_results = validation_results

    @property
    def estimator_(self):
        """
        Alias for the trained classifier.

        Returns:
        Any: Trained classifier.
        """
        return self.classifier.classifier


class Classifier(BaseEstimator, ClassifierMixin):
    """
    Wrapper class for classifiers with added functionality.

    Parameters:
    - classifier: Base classifier or string identifier for XGBoost.
    - shap_algorithm (str): SHAP algorithm to use.
    - use_sample_weight (bool): Whether to use sample weights during training.

    Methods:
    - fit(X, y): Fit the classifier to the data.
    - predict(X): Predict labels for input data.
    - cross_validation(X, y, k_folds): Perform cross-validation and store scores.
    - interpret(X): Interpret the model using SHAP values.
    - feature_importances_: Get feature importances based on SHAP values.

    """

    X: np.array
    y: np.array
    shap_values: list
    shap_algorithm: str
    mean_importance: np.ndarray
    use_sample_weight: bool

    def __init__(
        self, classifier, shap_algorithm: str = "auto", use_sample_weight: bool = True
    ):
        """
        Initialize the Classifier with a base classifier and optional parameters.

        Parameters:
        - classifier: Base classifier or string identifier for XGBoost.
        - shap_algorithm: SHAP algorithm to use.
        - use_sample_weight: Whether to use sample weights during training.

        Returns:
        None
        """
        if isinstance(classifier, str):
            if classifier == "xgboost":
                self.classifier = xgboost.XGBClassifier(
                    max_depth=30,
                    eval_metric="logloss",
                    verbosity=0,
                )
        else:
            fit_method = getattr(classifier, "fit", None)
            predict_method = getattr(classifier, "predict", None)
            if callable(fit_method) and callable(predict_method):
                self.classifier = classifier
            else:
                raise ValueError(
                    "The classifier does not have a fit and/or predict method"
                )
        self.shap_algorithm = shap_algorithm
        self.use_sample_weight = use_sample_weight

    def fit(self, X, y):
        """
        Fit the classifier to the data.

        Parameters:
        - X: Input data.
        - y: Target labels.

        Returns:
        Classifier: The fitted classifier.
        """
        self.X = X
        self.y = y
        if self.use_sample_weight:
            sample_weights = class_weight.compute_sample_weight(
                class_weight="balanced", y=y
            )
            self.classifier.fit(X, y, sample_weight=sample_weights)

        else:
            self.classifier.fit(X, y)
        return self

    def predict(self, X):
        """
        Predict labels for input data.

        Parameters:
        - X: Input data.

        Returns:
        np.array: Predicted labels.
        """

        return self.classifier.predict(X)

    def cross_validation(self, X, y, k_folds: int = 5):
        """
        Perform cross-validation and store scores.

        Parameters:
        - X: Input data.
        - y: Target labels.
        - k_folds: Number of folds for cross-validation.

        Returns:
        None
        """
        self.scores = cross_val_score(self.classifier, X, y, cv=k_folds)

    def interpret(self, X):
        """
        Interpret the model using SHAP values.

        Parameters:
        - X: Input data.

        Returns:
        None
        """
        self.X = X

        if self.shap_algorithm == "permutation":
            explainer = shap.Explainer(
                self.classifier.predict, X, algorithm=self.shap_algorithm
            )

            self.shap_values = explainer(X, max_evals=2 * X.shape[1] + 1).values

        elif (
            self.shap_algorithm == "tree"
            or self.shap_algorithm == "auto"
            or self.shap_algorithm == "partition"
        ):
            explainer = shap.Explainer(
                self.classifier, X, algorithm=self.shap_algorithm
            )
            self.shap_values = explainer.shap_values(X)
        elif self.shap_algorithm == "linear":
            explainer = shap.KernelExplainer(
                self.classifier.predict, data=X, algorithm=self.shap_algorithm
            )
            self.shap_values = explainer.shap_values(X)

        elif self.shap_algorithm == "predict_proba":
            explainer = shap.KernelExplainer(
                self.classifier.predict_proba,
                data=shap.sample(X, 20),
                algorithm=self.shap_algorithm,
            )
            self.shap_values = explainer.shap_values(X)

        if isinstance(self.shap_values, list):
            self.shap_values = np.swapaxes(np.array(self.shap_values), 1, 2)

        self.explainer = explainer
        self.mean_importance = np.mean(abs(self.shap_values), axis=0)

    @property
    def feature_importances_(self):
        """
        Get feature importances based on SHAP values.

        Returns:
        np.array: Feature importances.
        """
        self.interpret(self.X)
        return self.mean_importance

    @property
    def classes_(self):
        return np.unique(self.y)


def encode_labels(labels: np.ndarray) -> np.ndarray:
    """
    Encode labels using LabelEncoder.

    Parameters:
    - labels: Array of labels.

    Returns:
    np.ndarray: Encoded labels.
    """
    encoder = LabelEncoder()

    return encoder.fit_transform(labels)


def format_data(quant_matrix: QuantMatrix) -> np.ndarray:
    """
    Format quantitative matrix data.

    Parameters:
    - quant_matrix: QuantMatrix instance.

    Returns:
    np.ndarray: Formatted data.
    """
    X = quant_matrix.quantitative_data.X.copy().transpose()

    return np.nan_to_num(X, copy=True, nan=0.0)



"""quant_matrix module

instanciate a quant matrix:

>>> from dpks.quant_matrix import QuantMatrix
>>> quant_matrix = QuantMatrix( quantification_file="tests/input_files/minimal_matrix.tsv", design_matrix_file="tests/input_files/minimal_design_matrix.tsv")

"""
from __future__ import annotations

from typing import Union, List, Any, Tuple, Optional

import time

import anndata as ad
import gseapy as gp

import xgboost
from imblearn.under_sampling import RandomUnderSampler
from pandas import Series, DataFrame
from sklearn.model_selection import cross_val_score, StratifiedKFold
from unipressed import IdMappingClient

from dpks.param_search import GeneticAlgorithmSearch, RandomizedSearch, ParamSearchResult  # type: ignore
import matplotlib
import numpy as np
import pandas as pd  # type: ignore
from sklearn.preprocessing import StandardScaler, LabelEncoder

from dpks.annotate_proteins import get_protein_labels
from dpks.classification import Classifier, encode_labels, format_data, TrainResult
from dpks.differential_testing import DifferentialTest
from dpks.imputer import (
    ImputerMethod,
    UniformRangeImputer,
    UniformPercentileImputer,
)
from dpks.normalization import (
    TicNormalization,
    MedianNormalization,
    MeanNormalization,
    Log2Normalization,
    NormalizationMethod,
    RTSlidingWindowNormalization,
)
from dpks.parsers import parse_diann
from dpks.plot import SHAPPlot, RFEPCA
from dpks.quantification import TopN, MaxLFQ
from dpks.scaling import (
    ScalingMethod,
    ZScoreScaling,
    MinMaxScaling,
    AbsMaxScaling,
)

from dpks.interpretation import BootstrapInterpreter


class QuantMatrix:
    """Class for working with quantitative matrices."""
    quantification_file_path: Union[str, pd.DataFrame]
    design_matrix_file: Union[str, pd.DataFrame]
    num_rows: int
    num_samples: int
    quantitative_data: ad.AnnData
    explain_results: Optional[list[tuple[Any, BootstrapInterpreter]]]

    def __init__(
        self,
        quantification_file: Union[str, pd.DataFrame],
        design_matrix_file: Union[str, pd.DataFrame],
        annotation_fasta_file: str = None,
        quant_type: str = "gps",
        diann_qvalue: float = 0.01,
    ) -> None:
        """Initialize the QuantMatrix instance.

        Args:
            quantification_file (Union[str, pd.DataFrame]): Path to the quantification file or DataFrame.
            design_matrix_file (Union[str, pd.DataFrame]): Path to the design matrix file or DataFrame.
            annotation_fasta_file (str, optional): Path to the annotation FASTA file. Defaults to None.
            quant_type (str, optional): Type of quantification. Defaults to "gps".
            diann_qvalue (float, optional): DIANN q-value. Defaults to 0.01.

        Examples:
            >>> quant_matrix = QuantMatrix("quantification.tsv", "design_matrix.csv", annotation_fasta_file="annotation.fasta")
        """

        self.annotated = False
        self.explain_results = None
        if isinstance(design_matrix_file, str):
            design_matrix_file = pd.read_csv(design_matrix_file, sep="\t")

            design_matrix_file.columns = map(str.lower, design_matrix_file.columns)

        if isinstance(quantification_file, str):
            if quant_type == "gps":
                quantification_file = pd.read_csv(quantification_file, sep="\t")

            elif quant_type == "diann":
                quantification_file = parse_diann(quantification_file, diann_qvalue)

        else:
            if quant_type == "diann":
                quantification_file = parse_diann(quantification_file, diann_qvalue)

        self.num_samples = len(design_matrix_file)
        self.num_rows = len(quantification_file)

        rt_column = ""

        if "RT" in quantification_file:
            rt_column = "RT"

        elif "RetentionTime" in quantification_file:
            rt_column = "RetentionTime"

        if rt_column:
            quantification_file = quantification_file.sort_values(rt_column)

        quantification_file = quantification_file.reset_index(drop=True)

        quantitative_data = (
            quantification_file[list(design_matrix_file["sample"])]
            .copy()
            .set_index(np.arange(self.num_rows, dtype=int).astype(str))
        )

        row_obs = quantification_file.drop(
            list(design_matrix_file["sample"]), axis=1
        ).set_index(np.arange(self.num_rows, dtype=int).astype(str))

        if annotation_fasta_file is not None:
            row_obs["ProteinLabel"] = get_protein_labels(
                row_obs["Protein"], annotation_fasta_file
            )

        self.quantitative_data = ad.AnnData(
            quantitative_data,
            obs=row_obs,
            var=design_matrix_file.copy().set_index(design_matrix_file["sample"]),
        )

    @property
    def proteins(self) -> List[str]:

        return list(self.quantitative_data.obs["Protein"].unique())

    @property
    def protein_labels(self) -> List[str]:

        return self.row_annotations["ProteinLabel"].to_list()

    @property
    def sample_groups(self) -> List[str]:

        return self.sample_annotations["group"].to_list()

    @property
    def peptides(self) -> List[str]:

        return list(self.quantitative_data.obs["PeptideSequence"].unique())

    @property
    def precursors(self) -> List[str]:

        self.row_annotations["PrecursorId"] = (
            self.row_annotations["PeptideSequence"]
            + "_"
            + self.row_annotations["Charge"].astype(str)
        )

        return list(self.row_annotations["PrecursorId"].unique())

    @property
    def sample_annotations(self) -> pd.DataFrame:

        return self.quantitative_data.var

    @property
    def row_annotations(self) -> pd.DataFrame:

        return self.quantitative_data.obs

    @row_annotations.setter
    def row_annotations(self, value: pd.DataFrame) -> None:

        self.quantitative_data.obs = value

    def get_samples(self, group: int) -> List[str]:

        return list(
            self.sample_annotations[self.sample_annotations["group"] == group]["sample"]
        )

    def get_pairs(self, samples: list) -> List[str]:

        sorted_samples = (
            self.sample_annotations[self.sample_annotations["sample"].isin(samples)]
            .set_index("sample")
            .loc[samples]
        )

        return list(sorted_samples["pair"])

    def filter(
        self,
        peptide_q_value: float = 0.01,
        protein_q_value: float = 0.01,
        remove_decoys: bool = True,
        remove_contaminants: bool = True,
        remove_non_proteotypic: bool = True,
    ) -> QuantMatrix:
        """Filter the QuantMatrix.

        Args:
            peptide_q_value (float, optional): Peptide q-value threshold. Defaults to 0.01.
            protein_q_value (float, optional): Protein q-value threshold. Defaults to 0.01.
            remove_decoys (bool, optional): Whether to remove decoy entries. Defaults to True.
            remove_contaminants (bool, optional): Whether to remove contaminant entries. Defaults to True.
            remove_non_proteotypic (bool, optional): Whether to remove non-proteotypic entries. Defaults to True.

        Returns:
            QuantMatrix: Filtered QuantMatrix object.

        Examples:
            >>> print(quant_matrix.to_df().shape)
            (16679, 26)
            >>> print(quant_matrix.filter(peptide_q_value=0.001).to_df().shape)
            (15355, 26)

        """

        filtered_data = self.quantitative_data

        if "PeptideQValue" in self.quantitative_data.obs:
            filtered_data = self.quantitative_data[
                (self.quantitative_data.obs["PeptideQValue"] <= peptide_q_value)
            ].copy()

        if "ProteinQValue" in self.quantitative_data.obs:
            filtered_data = self.quantitative_data[
                (self.quantitative_data.obs["ProteinQValue"] <= protein_q_value)
            ].copy()

        if remove_decoys:
            if "Decoy" in filtered_data.obs:
                filtered_data = filtered_data[filtered_data.obs["Decoy"] == 0].copy()

        if remove_contaminants:
            filtered_data = filtered_data[
                ~filtered_data.obs["Protein"].str.contains("contam")
            ].copy()

            filtered_data = filtered_data[
                ~filtered_data.obs["Protein"].str.contains("cont_crap")
            ].copy()

        if remove_non_proteotypic:
            filtered_data = filtered_data[
                ~filtered_data.obs["Protein"].str.contains(";")
            ].copy()

        self.num_rows = len(filtered_data)

        quantitative_data = (
            filtered_data.to_df()[list(filtered_data.var["sample"])]
            .copy()
            .set_index(np.arange(self.num_rows, dtype=int).astype(str))
        )

        row_obs = filtered_data.obs.set_index(
            np.arange(self.num_rows, dtype=int).astype(str)
        )

        self.quantitative_data = ad.AnnData(
            quantitative_data,
            obs=row_obs,
            var=filtered_data.var
        )

        return self

    def scale(
        self,
        method: str,
    ) -> QuantMatrix:
        """Scale the QuantMatrix data at the feature level (i.e Precursor or Protein).

        Args:
            method (str): Scaling method. Options are 'zscore', 'minmax', or 'absmax'.

        Returns:
            QuantMatrix: Scaled QuantMatrix object.

        Raises:
            ValueError: If the provided scaling method is not supported.

        """
        base_method: ScalingMethod = ScalingMethod()

        if method == "zscore":
            base_method = ZScoreScaling()

        elif method == "minmax":
            base_method = MinMaxScaling()

        elif method == "absmax":
            base_method = AbsMaxScaling()

        else:

            raise ValueError(f"Unsupported scaling method: {method}")

        self.quantitative_data.X = base_method.fit_transform(self.quantitative_data.X)

        return self

    def normalize(
        self,
        method: str,
        log_transform: bool = True,
        use_rt_sliding_window_filter: bool = False,
        **kwargs: Union[int, bool, str],
    ) -> QuantMatrix:
        """Normalize the QuantMatrix data.

        Args:
            method (str): Normalization method. Options are 'tic', 'median', or 'mean'.
            log_transform (bool, optional): Whether to log-transform the data. Defaults to True.
            use_rt_sliding_window_filter (bool, optional): Whether to use a sliding window filter. Defaults to False. Can only use if a RetentionTime column was loaded in the QuantMatrix
            **kwargs: Additional keyword arguments depending on the chosen method.

        Returns:
            QuantMatrix: Normalized QuantMatrix object.

        Raises:
            ValueError: If the provided normalization method is not supported.

        Examples:
            >>> quant_matrix.normalize(method="mean")

        """

        base_method: NormalizationMethod = NormalizationMethod()

        if method == "tic":
            base_method = TicNormalization()

        elif method == "median":
            base_method = MedianNormalization()

        elif method == "mean":
            base_method = MeanNormalization()

        else:
            raise ValueError(f"Unsupported normalization method: {method}")

        if use_rt_sliding_window_filter:
            minimum_data_points = int(kwargs.get("minimum_data_points", 100))
            stride = int(kwargs.get("stride", 1))
            use_overlapping_windows = bool(kwargs.get("use_overlapping_windows", True))
            rt_unit = str(kwargs.get("rt_unit", "minute"))

            rt_window_normalization = RTSlidingWindowNormalization(
                base_method=base_method,
                minimum_data_points=minimum_data_points,
                stride=stride,
                use_overlapping_windows=use_overlapping_windows,
                rt_unit=rt_unit,
            )

            self.quantitative_data.X = rt_window_normalization.fit_transform(self)

        else:
            self.quantitative_data.X = base_method.fit_transform(
                self.quantitative_data.X
            )

        if log_transform:
            self.quantitative_data.X = Log2Normalization().fit_transform(
                self.quantitative_data.X
            )

        return self

    def quantify(
        self,
        method: str,
        **kwargs: Union[int, str],
    ) -> QuantMatrix:
        """Calculate protein quantities.

        Args:
            method (str): Quantification method. Options are 'top_n' or 'maxlfq'.
            **kwargs: Additional keyword arguments depending on the chosen method.

        Returns:
            QuantMatrix: Quantified protein matrix.

        Raises:
            ValueError: If the provided quantification method is not supported.

        Examples:
            >>> quant_matrix.quantify(method="top_n", top_n=1)

        """

        if method == "top_n":
            level = str(kwargs.get("level", "protein"))
            top_n = int(kwargs.get("top_n", 1))
            summarization_method = str(kwargs.get("summarization_method", "sum"))

            quantifications = TopN(
                top_n=top_n, level=level, summarization_method=summarization_method
            ).quantify(self)

            design_matrix = self.quantitative_data.var

            protein_quantifications = QuantMatrix(
                quantifications, design_matrix_file=design_matrix
            )

        elif method == "maxlfq":
            level = str(kwargs.get("level", "protein"))
            threads = int(kwargs.get("threads", 1))
            minimum_subgroups = int(kwargs.get("minimum_subgroups", 1))
            top_n = int(kwargs.get("top_n", 0))

            quantifications = MaxLFQ(
                level=level,
                threads=threads,
                minimum_subgroups=minimum_subgroups,
                top_n=top_n,
            ).quantify(self)

            design_matrix = self.quantitative_data.var

            protein_quantifications = QuantMatrix(
                quantifications, design_matrix_file=design_matrix
            )

        else:
            raise ValueError(f"Unsupported quantification method: {method}")

        return protein_quantifications

    def impute(self, method: str, **kwargs: int) -> QuantMatrix:
        """Impute missing values in the quantitative data.

        Args:
            method (str): The imputation method to use. Options are "uniform_percentile" and "uniform_range"
            **kwargs (int): Additional keyword arguments specific to the imputation method.

        Returns:
            QuantMatrix: The QuantMatrix object with missing values imputed.

        Raises:
            ValueError: If an unsupported imputation method is provided.

        Examples:
            >>> quant_matrix.impute(method="uniform_percentile", percentile=0.1)

        """

        base_method: ImputerMethod = ImputerMethod()

        if method == "uniform_percentile":
            percentile = float(kwargs.get("percentile", 0.1))

            base_method = UniformPercentileImputer(percentile=percentile)

        elif method == "uniform_range":
            maxvalue = int(kwargs.get("maxvalue", 1))
            minvalue = int(kwargs.get("minvalue", 0))

            base_method = UniformRangeImputer(maxvalue=maxvalue, minvalue=minvalue)

        else:

            raise ValueError(f"Unsupported imputation method: {method}")

        self.quantitative_data.X = base_method.fit_transform(self.quantitative_data.X)

        return self

    def compare(
        self,
        method: str,
        comparisons: list,
        min_samples_per_group: int = 2,
        level: str = "protein",
        multiple_testing_correction_method: str = "fdr_tsbh",
    ) -> QuantMatrix:
        """Compare groups by differential testing.

        Args:
            method (str): Statistical comparison method. Options are 'ttest', 'linregress', 'anova', 'ttest_paired'.
            comparisons (list): List of tuples specifying the group comparisons.
            min_samples_per_group (int, optional): Minimum number of samples per group. Defaults to 2.
            level (str, optional): Level of comparison. Defaults to 'protein'.
            multiple_testing_correction_method (str, optional): Method for multiple testing correction. Defaults to 'fdr_tsbh'.

        Returns:
            QuantMatrix: Matrix containing the results of the differential testing.

        Raises:
            ValueError: If the provided statistical comparison method is not supported.

        Examples:
            >>> quantified_data = quantified_data.compare(
            >>>     method="linregress",
            >>>     min_samples_per_group=2,
            >>>     comparisons=[(2, 1), (3, 1)]
            >>> )


        """

        if not method in {
            'ttest',
            'linregress',
            'anova',
            'ttest_paired'
        }:
            raise ValueError(f"Unsupported statistical comparison method: {method}")

        differential_test = DifferentialTest(
            method,
            comparisons,
            min_samples_per_group,
            level,
            multiple_testing_correction_method,
        )

        compared_data = differential_test.test(self)

        self.row_annotations = compared_data.row_annotations.copy()

        return self

    def explain(
        self,
        clf,
        comparisons: list,
        n_iterations: int = 100,
        downsample_background: bool = True,
        feature_column: str = "Protein",
    ) -> QuantMatrix:
        """Explain group differences using explainable machine learning and feature importance.

        Args:
            clf: Classifier object used for prediction.
            comparisons (list): List of tuples specifying the group comparisons.
            n_iterations (int, optional): Number of iterations for bootstrapping. Defaults to 100.
            downsample_background (bool, optional): Whether to downsample the background. Defaults to True.
            feature_column (str, optional): Name of the feature column. Defaults to 'Protein'.

        Returns:
            QuantMatrix: Matrix containing the results of the explanation.

        Examples:
            >>> import xgboost
            >>>
            >>> clf = xgboost.XGBClassifier(
            >>>     max_depth=2,
            >>>     reg_lambda=2,
            >>>     objective="binary:logistic",
            >>>     seed=42
            >>> )
            >>>
            >>> quantified_data = quantified_data.explain(
            >>>     clf,
            >>>     comparisons=[(2, 1), (3, 1)],
            >>>     n_iterations=10,
            >>>     downsample_background=True
            >>> )

        """
        explain_results = []

        for comparison in comparisons:
            X, y = self.to_ml(feature_column=feature_column, comparison=comparison)

            scaler = StandardScaler()

            X[:] = scaler.fit_transform(X[:])

            clf_ = Classifier(clf)

            interpreter = BootstrapInterpreter(
                feature_names=X.columns,
                n_iterations=n_iterations,
                downsample_background=downsample_background,
            )

            interpreter.fit(X, y, clf_)

            explain_results.append((comparison, interpreter))

            importances_df = interpreter.importances[
                ["feature", "mean_shap", "mean_rank"]
            ].set_index("feature")

            importances_df = importances_df.rename(
                columns={
                    "mean_shap": f"MeanSHAP{comparison[0]}-{comparison[1]}",
                    "mean_rank": f"MeanRank{comparison[0]}-{comparison[1]}",
                }
            )

            self.row_annotations = self.row_annotations.join(
                importances_df, on="Protein"
            )

        self.explain_results = explain_results

        return self

    def enrich(
        self,
        method: str = "overreptest",
        libraries: Optional[list[str]] = None,
        organism: str = "human",
        background: Optional[Union[list[str], str]] = None,
        filter_pvalue: bool = False,
        pvalue_cutoff: float = 0.1,
        pvalue_column: str = "CorrectedPValue2-1",
        filter_shap: bool = False,
        shap_cutoff: float = 0.0,
        shap_column: str = "MeanSHAP2-1",
        subset_library: bool = False,
    ):
        """Perform gene set enrichment analysis.

        Args:
           method (str, optional): Enrichment method to use. Options are "enrichr_overreptest" and "overreptest". Defaults to "overreptest".
           libraries (Optional[List[str]], optional): List of gene set libraries. Defaults to None.
           organism (str, optional): Organism for the analysis. Defaults to "human".
           background (Optional[Union[List[str], str]], optional): Background gene set. Defaults to None.
           filter_pvalue (bool, optional): Whether to filter by p-value. Defaults to False.
           pvalue_cutoff (float, optional): P-value cutoff for filtering. Defaults to 0.1.
           pvalue_column (str, optional): Column name for p-values. Defaults to "CorrectedPValue2-1".
           filter_shap (bool, optional): Whether to filter by SHAP value. Defaults to False.
           shap_cutoff (float, optional): SHAP value cutoff for filtering. Defaults to 0.0.
           shap_column (str, optional): Column name for SHAP values. Defaults to "MeanSHAP2-1".
           subset_library (bool, optional): Whether to subset the library. Defaults to False.

        Returns:
           Any: Enrichment result.

        Raises:
           ValueError: If the method is not supported.

        Examples:
            >>> enr = quantified_data.enrich(
            >>>     method="enrichr_overreptest",
            >>>     filter_pvalue=True,
            >>>     pvalue_column="CorrectedPValue2-1",
            >>>     pvalue_cutoff=0.1
            >>> )

        """
        if not self.annotated:
            self.annotate()

        if not libraries:
            libraries = ["GO_Biological_Process_2023"]

        gene_df = pd.DataFrame()

        if filter_pvalue:
            gene_df = self.row_annotations[
                self.row_annotations[pvalue_column] < pvalue_cutoff
            ]

        if filter_shap:
            gene_df = self.row_annotations[
                self.row_annotations[shap_column] > shap_cutoff
            ]

        genes = gene_df["Gene"].to_list()

        if subset_library:
            temp_libraries = []

            for library in libraries:
                go_bp = gp.get_library(name=library, organism=organism)

                gene_set = set(gene_df["Gene"].to_list())

                bio_process_subset = dict()

                for key, value in go_bp.items():
                    for gene in value:
                        if gene in gene_set:
                            bio_process_subset[key] = value

                temp_libraries.append(bio_process_subset)

            libraries = temp_libraries

        enr = None

        if method == "overreptest":
            if background:
                enr = gp.enrich(
                    gene_list=genes,
                    gene_sets=libraries,
                    background=background,
                )

            else:
                enr = gp.enrich(gene_list=genes, gene_sets=libraries)

        elif method == "enrichr_overreptest":
            if background:
                enr = gp.enrichr(
                    gene_list=genes,
                    gene_sets=libraries,
                    organism=organism,
                    background=background,
                )

            else:
                enr = gp.enrichr(
                    gene_list=genes,
                    gene_sets=libraries,
                    organism=organism,
                )

        else:

            raise ValueError(f"Unsupported pathway enrichment method: {method}")

        return enr

    def annotate(self):
        """Annotate proteins with gene names.

        Returns:
            QuantMatrix: The annotated QuantMatrix object.

        Examples:
            >>> quant_matrix.annotate()

        """
        request = IdMappingClient.submit(
            source="UniProtKB_AC-ID", dest="Gene_Name", ids=self.proteins
        )

        while True:
            status = request.get_status()
            if status in {"FINISHED", "ERROR"}:
                break
            else:
                time.sleep(1)

        translation_result = list(request.each_result())

        id_mapping = dict()

        for id_result in translation_result:
            mapping = id_mapping.get(id_result["from"], [])

            mapping.append(id_result["to"])

            id_mapping[id_result["from"]] = mapping

        final_mapping = dict()

        for key, value in id_mapping.items():
            value = value[0]

            final_mapping[key] = value

        mapping_df = pd.DataFrame(
            {"Protein": final_mapping.keys(), "Gene": final_mapping.values()}
        )

        self.row_annotations = self.row_annotations.join(
            mapping_df.set_index("Protein"), on="Protein", how="left"
        )

        self.row_annotations["Gene"] = self.row_annotations["Gene"].fillna(
            self.row_annotations["Protein"]
        )

        self.annotated = True

        return self

    def predict(
        self,
        classifier,
        scaler: Any = None,
        scale: bool = True,
    ) -> QuantMatrix:
        """Predict labels using a classifier.

        Args:
            classifier: The classifier model to use for prediction.
            scaler (optional): The scaler object to use for data scaling.
            scale (bool): Whether to scale the data before prediction. Defaults to True.

        Returns:
            QuantMatrix: The QuantMatrix object with predicted labels.

        Examples:
            >>> quant_matrix.predict(classifier=clf, scaler=std_scaler)

        """
        X = format_data(self)

        if scale:
            if scaler:
                X = scaler.transform(X)
            else:
                scaler = StandardScaler()
                X = scaler.fit_transform(X)

        classifier = Classifier(classifier=classifier)

        self.sample_annotations["Prediction"] = classifier.predict(X)

        return self

    def interpret(
        self,
        classifier,
        scaler: Any = None,
        shap_algorithm: str = "auto",
        scale: bool = True,
        downsample_background=False,
    ) -> QuantMatrix:
        """Interpret the model's predictions using SHAP values.

        Args:
            classifier: The classifier model to interpret.
            scaler (optional): The scaler object to use for data scaling.
            shap_algorithm (str): The SHAP algorithm to use. Defaults to "auto".
            scale (bool): Whether to scale the data before interpretation. Defaults to True.
            downsample_background (bool): Whether to downsample background data. Defaults to False.

        Returns:
            QuantMatrix: The QuantMatrix object with SHAP values added to observations.

        Examples:
            >>> quant_matrix.interpret(classifier=clf, scaler=std_scaler)

        """
        X = format_data(self)
        y = encode_labels(self.quantitative_data.var["group"].values)

        if scale:
            if scaler:
                X = scaler.transform(X)
            else:
                scaler = StandardScaler()
                X = scaler.fit_transform(X)

        classifier = Classifier(classifier=classifier, shap_algorithm=shap_algorithm)

        if downsample_background:
            rus = RandomUnderSampler(random_state=0)
            X_resampled, y_resampled = rus.fit_resample(X, y)
            classifier.interpret(X_resampled)
            self.transformed_data = X_resampled
            self.y_resampled = y_resampled
        else:
            classifier.interpret(X)
            self.transformed_data = X

        self.classifier = classifier
        shap_values = classifier.feature_importances_.tolist()

        self.quantitative_data.obs["SHAP"] = shap_values

        self.shap = classifier.shap_values

        return self

    def train(
        self,
        classifier,
        scaler: Any = None,
        scale: bool = True,
        validate: bool = True,
        scoring: str = "accuracy",
        num_folds: int = 3,
        random_state: int = 42,
        shuffle: bool = False,
    ) -> TrainResult:
        """Train a classifier on the quantitative data.

        Args:
            classifier: The classifier object or class to use for training.
            scaler (Any): The scaler object to scale the data. Defaults to None.
            scale (bool): Whether to scale the data. Defaults to True.
            validate (bool): Whether to perform cross-validation. Defaults to True.
            scoring (str): The scoring metric for cross-validation. Defaults to "accuracy".
            num_folds (int): The number of folds for cross-validation. Defaults to 3.
            random_state (int): Random seed for reproducibility. Defaults to 42.
            shuffle (bool): Whether to shuffle the data before splitting in cross-validation. Defaults to False.

        Returns:
            TrainResult: The result of the training process, including the trained classifier, scaler, and validation scores.

        Examples:
            >>> result = quant_matrix.train(classifier=RandomForestClassifier(), validate=True)

        """
        X = format_data(self)
        y = encode_labels(self.quantitative_data.var["group"].values)

        if scale:
            if scaler:
                X = scaler.transform(X)
            else:
                scaler = StandardScaler()
                X = scaler.fit_transform(X)

        classifier = Classifier(classifier=classifier)

        validation_result = np.array([])

        if validate:
            cv = StratifiedKFold(num_folds, shuffle=shuffle, random_state=random_state)
            validation_result = cross_val_score(
                classifier, X, y, scoring=scoring, cv=cv
            )

        classifier.fit(X, y)

        return TrainResult(classifier, scaler, validation_result)

    def optimize(
        self,
        classifier,
        param_search_method: str,
        param_grid: dict,
        scaler: Any = None,
        scale: bool = True,
        threads: int = 1,
        random_state: int = 42,
        folds: int = 3,
        verbose: Union[bool, int] = False,
        **kwargs: Union[dict, int, str, bool],
    ) -> ParamSearchResult:
        """Optimize hyperparameters of a classifier using different search methods.

        Args:
            classifier: The classifier object or class to optimize.
            param_search_method (str): The parameter search method to use ("genetic" or "random").
            param_grid (dict): The parameter grid to search over.
            scaler (Any): The scaler object to scale the data. Defaults to None.
            scale (bool): Whether to scale the data. Defaults to True.
            threads (int): The number of threads to use for optimization. Defaults to 1.
            random_state (int): Random seed for reproducibility. Defaults to 42.
            folds (int): The number of folds for cross-validation. Defaults to 3.
            verbose (Union[bool, int]): Verbosity level. Defaults to False.
            **kwargs: Additional keyword arguments specific to each search method.

        Returns:
            ParamSearchResult: The result of the parameter search, including the best estimator and parameter populations.

        Examples:
            >>> param_grid = {'max_depth': [3, 5, 7], 'min_samples_split': [2, 5, 10]}
            >>> result = quant_matrix.optimize(classifier=DecisionTreeClassifier(), param_search_method='random', param_grid=param_grid, verbose=True)
            >>> result.best_estimator_
            DecisionTreeClassifier(max_depth=5, min_samples_split=10)
        """
        X = format_data(self)
        y = encode_labels(self.quantitative_data.var["group"].values)

        if scale:
            if scaler:
                X = scaler.transform(X)
            else:
                scaler = StandardScaler()
                X = scaler.fit_transform(X)

        result = None

        if param_search_method == "genetic":
            gas = GeneticAlgorithmSearch(
                classifier,
                param_grid=param_grid,
                threads=threads,
                folds=folds,
                n_survive=kwargs.get("n_survive", 5),
                pop_size=kwargs.get("pop_size", 10),
                n_generations=kwargs.get("n_generations", 20),
                verbose=verbose,
                random_state=kwargs.get("random_state", None),
                shuffle=kwargs.get("shuffle", False),
            )
            parameter_populations = gas.fit(X, y)

            result = ParamSearchResult(
                classifier=gas.best_estimator_,
                result=parameter_populations,
            )

        elif param_search_method == "random":
            randomized_search = RandomizedSearch(
                classifier,
                param_grid=param_grid,
                folds=folds,
                random_state=random_state,
                n_iter=kwargs.get("n_iter", 30),
                n_jobs=threads,
                scoring=kwargs.get("scoring", "accuracy"),
                verbose=verbose,
            )

            result = randomized_search.fit(X, y)

        return result



    def plot(
        self,
        plot_type: str,
        save: bool = False,
        fig: matplotlib.figure.Figure = None,
        ax: Union[list, matplotlib.axes.Axes] = None,
        **kwargs: Union[
            np.ndarray,
            int,
            list,
            str,
        ],
    ) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        """Generate plots based on specified plot type.

        Args:
            plot_type (str): The type of plot to generate. Possible values are:
                - "shap_summary": SHAP summary plot.
                - "rfe_pca": Recursive Feature Elimination (RFE) with Principal Component Analysis (PCA) plot.
            save (bool): Whether to save the plot. Defaults to False.
            fig (matplotlib.figure.Figure): The matplotlib figure object. Defaults to None.
            ax (Union[list, matplotlib.axes.Axes]): The list of matplotlib axes objects or a single axes object. Defaults to None.
            **kwargs: Additional keyword arguments specific to each plot type.

        Returns:
            tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]: The matplotlib figure and axes objects.

        Raises:
            ValueError: If an unsupported plot type is provided.

        Examples:
            >>> fig, ax = quant_matrix.plot(plot_type='shap_summary', save=True, n_display=10)
        """

        if plot_type == "shap_summary":
            try:
                getattr(self, "shap")
            except AttributeError:
                print("SHAP values have not been generated")
            cmap = kwargs.get(
                "cmap",
                [
                    "#ff4800",
                    "#ff4040",
                    "#a836ff",
                    "#405cff",
                    "#05c9fa",
                ],
            )

            order_by = kwargs.get("order_by", "shap")

            fig, ax = SHAPPlot(
                fig=fig,
                ax=ax,
                shap_values=self.shap,
                X=self.transformed_data,
                qm=self,
                cmap=cmap,
                n_display=kwargs.get("n_display", 5),
                jitter=kwargs.get("jitter", 0.1),
                alpha=kwargs.get("alpha", 0.75),
                n_bins=kwargs.get("n_bins", 100),
                feature_column=kwargs.get("feature_column", "Protein"),
                order_by=order_by,
            ).plot()

        if plot_type == "rfe_pca":
            cmap = kwargs.get("cmap", "coolwarm")
            cutoffs = list(kwargs.get("cutoffs", [100, 50, 10]))
            fig, ax = RFEPCA(
                fig=fig, axs=ax, qm=self, cutoffs=cutoffs, cmap=cmap
            ).plot()

        if save:
            filepath = str(kwargs.get("filepath", f"{plot_type}.png"))
            dpi = int(kwargs.get("dpi", 300))
            matplotlib.pyplot.savefig(filepath, dpi=dpi)

        return fig, ax

    def detect(self) -> None:
        """Not implemented

        Detect outliers in the samples
        """

        pass

    def write(self, file_path: str) -> None:
        """Write the QuantMatrix to a tab-separated file.

        Args:
            file_path (str): The path where the file will be saved.

        Returns:
            None

        Examples:
            >>> filename = "protein.tsv"
            >>> quant_matrix.write(filename)
        """

        self.to_df().to_csv(file_path, sep="\t", index=False)

    def to_df(self) -> pd.DataFrame:
        """Convert the QuantMatrix object to a pandas DataFrame.

        Returns:
            pd.DataFrame: DataFrame representation of the QuantMatrix.

        Examples:
            >>> quant_matrix.to_df()

        """

        quant_data = self.quantitative_data[self.row_annotations.index, :].to_df()

        merged = pd.concat([self.row_annotations, quant_data], axis=1)

        return merged
    def to_ml(
        self,
        feature_column: str = "Protein",
        label_column: str = "group",
        comparison: tuple = (1, 2),
    ) -> tuple[Any, Any]:
        """Converts the QuantMatrix object to features and labels for machine learning.

        Args:
            feature_column (str, optional): The column to use as features. Defaults to "Protein".
            label_column (str, optional): The column to use as labels. Defaults to "group".
            comparison (tuple, optional): The comparison groups. Defaults to (1, 2).

        Returns:
            tuple[Any, Any]: A tuple containing features and labels.

        Examples:
            >>> features, labels = quant_matrix.to_ml()
        """
        qm_df = self.to_df()

        samples = self.sample_annotations[
            self.sample_annotations["group"].isin(comparison)
        ]["sample"].to_list()

        transposed_features = qm_df.set_index(feature_column)[samples].T

        sample_annotations = self.sample_annotations.copy()

        sample_annotations_subset = sample_annotations[
            sample_annotations[label_column].isin(comparison)
        ].copy()

        encoder = LabelEncoder()

        sample_annotations_subset["label"] = encoder.fit_transform(
            sample_annotations_subset[label_column]
        )

        combined = transposed_features.join(
            sample_annotations_subset[["sample", "label"]].set_index("sample"),
            how="left",
        )

        return combined.loc[:, combined.columns != "label"], combined[["label"]]



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages  # type: ignore

version = "0.1.4"

install_requires = [
    "click",
    "numpy",
    "numba",
    "scipy",
    "networkx",
    "pandas",
    "seaborn",
    "anndata",
    "scikit-learn",
    "statsmodels",
    "biopython",
    "pytest",
    "pytest-runner",
    "xgboost",
    "shap",
    "imbalanced-learn",
    "kneed",
    "gseapy",
    "unipressed",
    "jupyterlab",
    "tbb"
]


setup(
    name="dpks",
    author="Aaron Scott",
    author_email="aaron.scott@med.lu.se",
    install_requires=install_requires,
    long_description="Data processing package for the statistical analysis and application of explainable machine learning for omics data.",
    include_package_data=True,
    packages=find_packages(include=["dpks", "dpks.*"]),
    url="https://github.com/InfectionMedicineProteomics/DPKS",
    version=version,
)



#!/usr/bin/env python3
"""Tests for the differential_testing module"""

import pytest
from dpks.quant_matrix import QuantMatrix


@pytest.mark.differential_testing()
def test_differential_testing(paths):
    """test the differential_testing object"""

    quant_matrix = QuantMatrix(
        quantification_file=str(paths["de_matrix_path"]),
        design_matrix_file=str(paths["de_design_matrix_path"]),
    )
    compared_data = (
        quant_matrix.filter()
        .normalize(
            method="mean",
            #                log_transform=True,
            #                use_rt_sliding_window_filter=True,
            #                window_length=100,
            #                stride=1
        )
        .quantify(method="top_n", top_n=1)
        .compare(
            method="linregress",
            group_a=4,
            group_b=6,
            min_samples_per_group=2,
            level="protein",
            multiple_testing_correction_method="fdr_tsbh",
        )
    )

    assert isinstance(compared_data, QuantMatrix)



#!/usr/bin/env python3
"""Tests for the plot module"""

import pytest
from dpks.plot import Plot


@pytest.mark.plot()
def test_plot(plot_object):
    """test the plot object"""

    assert isinstance(plot_object, Plot)



"""**impute quantitative matrices**, supports multiple methods as specified below"""
from typing import TYPE_CHECKING, Any

import numpy as np

if TYPE_CHECKING:
    from .quant_matrix import QuantMatrix
else:
    QuantMatrix = Any


class ImputerMethod:
    """the base class"""

    def __init__(self) -> None:
        """init"""

        pass

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """fit the transform"""

        pass


class UniformPercentileImputer(ImputerMethod):
    """uniform percentile imputer"""

    def __init__(self, percentile: float) -> None:
        """init"""
        self.percentile = percentile

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """fit the transform"""
        X = np.nan_to_num(X)  # Replace NaN with 0
        X_no_zero = X[X != 0]
        minvalue = X_no_zero.min()
        maxvalue = X_no_zero.max() * self.percentile + minvalue

        mask = X == 0
        c = np.count_nonzero(mask)
        nums = np.random.uniform(minvalue, maxvalue, c)
        X[mask] = nums

        return X


class UniformRangeImputer(ImputerMethod):
    """uniform imputer
    >>> from dpks.quant_matrix import QuantMatrix
    >>> quant_matrix = QuantMatrix( quantification_file="tests/input_files/minimal_matrix_missing.tsv", design_matrix_file="tests/input_files/minimal_design_matrix.tsv")
    >>> uniform_imputed_df = quant_matrix.impute(method="uniform", minvalue=1, maxvalue=2).to_df()[["PeptideSequence", "SAMPLE_1.osw", "SAMPLE_2.osw", "SAMPLE_3.osw"]]
    >>> uniform_imputed_df
        PeptideSequence  SAMPLE_1.osw  SAMPLE_2.osw  SAMPLE_3.osw
    0         EFMEEVIQR       69900.3           1.0      403947.0
    1  SSSGTPDLPVLLTDLK      115684.0      132524.0      217962.0
    2            PEPTIK           1.0       59295.7       24536.4
    """

    def __init__(self, minvalue: int, maxvalue: int) -> None:
        """init"""
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    def fit_transform(self, X: np.ndarray) -> np.ndarray:
        """fit the transform"""
        X = np.nan_to_num(X)  # Replace NaN with 0
        mask = X == 0
        c = np.count_nonzero(mask)
        nums = np.random.uniform(self.minvalue, self.maxvalue, c)

        X[mask] = nums

        return X
